# Copyright 2023-2026 The PULGON Project Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied. See the License for the specific language governing
# permissions and limitations under the License.

import copy
import itertools
import logging
from decimal import Decimal
from typing import Any, List, Optional, Tuple, Union

import ase
import cvxpy as cp
import numpy as np
import scipy as sp
import scipy.interpolate
import scipy.sparse as ss
import scipy.spatial.distance
from ase import Atoms
from pymatgen.core.operations import SymmOp
from pymatgen.util.coord import find_in_coord_list

from pulgon_tools.Irreps_tables import line_group_sympy
from pulgon_tools.Irreps_tables_withparities import (
    line_group_sympy_withparities,
)


def e() -> np.ndarray:
    """
    Returns: identity matrix
    """
    mat = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
    return mat


def Cn(n: Union[int, float]) -> np.ndarray:
    """
    Args:
        n: rotate 2*pi/n

    Returns: rotation matrix
    """
    mat = np.array(
        [
            [np.cos(2 * np.pi / n), -np.sin(2 * np.pi / n), 0],
            [np.sin(2 * np.pi / n), np.cos(2 * np.pi / n), 0],
            [0, 0, 1],
        ]
    )
    return mat


def sigmaV() -> np.ndarray:
    """

    Returns: mirror symmetric matrix

    """
    mat = np.array([[1, 0, 0], [0, -1, 0], [0, 0, 1]])
    return mat


def sigmaH() -> np.ndarray:
    """

    Returns: mirror symmetric matrix about x-y plane

    """
    mat = np.array([[1, 0, 0], [0, 1, 0], [0, 0, -1]])
    return mat


def U() -> np.ndarray:
    """

    Returns: A symmetric matrix about the x-axis

    """
    mat = np.array([[1, 0, 0], [0, -1, 0], [0, 0, -1]])
    return mat


def U_d(fid: Union[float, int]) -> np.ndarray:
    """

    Args:
        fid: the angle between symmetry axis d and
            axis x, d located in the x-y plane

    Returns: A symmetric matrix about the d-axis

    """
    mat = np.array(
        [
            [np.cos(2 * fid), np.sin(2 * fid), 0],
            [np.sin(2 * fid), -np.cos(2 * fid), 0],
            [0, 0, -1],
        ]
    )
    return mat


def S2n(n: Union[int, float]) -> np.ndarray:
    """
    Args:
        n: dihedral group, rotate 2*pi/n

    Returns: rotation and mirror matrix

    """
    mat = np.array(
        [
            [np.cos(np.pi / n), np.sin(np.pi / n), 0],
            [-np.sin(np.pi / n), np.cos(np.pi / n), 0],
            [0, 0, -1],
        ]
    )
    return mat


def sortrows(a: np.ndarray) -> np.ndarray:
    """
    :param a:
    :return: Compare each row in ascending order
    """
    return a[np.lexsort(np.rot90(a))]


def refine_cell(
    scale_pos: np.ndarray, numbers: np.ndarray, symprec: int = 4
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Refine fractional coordinates into [0,1) and remove duplicate atoms.

    Parameters
    ----------
    scale_pos : (N,3) or (3,) ndarray
        Fractional coordinates.
    numbers : (N,) ndarray
        Atomic species indices.
    symprec : int
        Decimal precision used to identify duplicates.

    Returns
    -------
    pos : ndarray
        Refined fractional coordinates.
    numbers : ndarray
        Corresponding atomic species.
    """

    scale_pos = np.asarray(scale_pos)

    # map coordinates to [0,1)
    scale_pos = np.mod(scale_pos, 1.0)

    # single atom case
    if scale_pos.ndim == 1:
        pos = np.round(scale_pos, symprec)
        return pos, numbers

    # round for numerical stability
    scale_pos = np.round(scale_pos, symprec)

    # remove duplicates while preserving order
    _, index = np.unique(scale_pos, axis=0, return_index=True)
    index = np.sort(index)

    pos = scale_pos[index]
    numbers = numbers[index]

    return pos, numbers


def frac_range(
    start: float,
    end: float,
    left: bool = True,
    right: bool = True,
    symprec: float = 0.01,
) -> list:
    """return the integer within the specified range

    Args:
        start: left boundary
        end: right boundary
        left: False means delete the left boundary
            element if it is an integer
        right: False means delete the right boundary
            element if it is an integer
        symprec: system precision

    Returns:

    """
    close = list(
        range(
            np.ceil(start).astype(np.int32), np.floor(end).astype(np.int32) + 1
        )
    )
    if left == False:
        if close[0] - start < symprec:
            close.pop(0)  # delete the left boundary
    if right == False:
        if close[-1] - end < symprec:
            close.pop()  # delete the right boundary
    return close


def decimal_places(x: float) -> int:
    """Return the number of decimal places of x
    using exact Decimal arithmetic."""
    d = Decimal(str(x))
    return abs(d.as_tuple().exponent)


def get_center_of_mass_periodic(atom: Atoms) -> np.ndarray:
    """Compute the center of mass in scaled coordinates
    with periodic wrapping.

    Uses circular mean for x and y; standard mean for z.

    Args:
        atom: ASE Atoms object.

    Returns:
        np.ndarray of scaled center-of-mass coordinates [cx, cy, cz].
    """
    L = np.array([1, 1, 1])
    x = atom.get_scaled_positions()
    theta = 2.0 * np.pi * x / L
    mass = atom.get_masses()
    mtheta = (
        np.arctan2(
            (-np.sin(theta) * np.expand_dims(mass, axis=1)).sum(axis=0)
            / len(theta),
            (-np.cos(theta) * np.expand_dims(mass, axis=1)).sum(axis=0)
            / len(theta),
        )
        + np.pi
    )
    center = L * mtheta / 2.0 / np.pi

    tmp = atom.get_center_of_mass(scaled=True)
    center[2] = tmp[2]
    return center


def find_axis_center_of_nanotube(atom: ase.atoms.Atoms) -> ase.atoms.Atoms:
    """remove the center of structure to (x,y):(0.5,0.5)
    Args:
        atom: initial structure

    Returns: centralized structure

    """
    n_st = atom.copy()
    center = get_center_of_mass_periodic(atom)

    pos = (
        # np.remainder(atom.get_scaled_positions() - center + 0.5, [1, 1, 1])
        np.remainder(
            atom.get_scaled_positions()
            # - [center[0],center[1],0] + [0.5, 0.5, 0],
            - center + [0.5, 0.5, 0],
            [1, 1, 1],
        )
        @ atom.cell
    )

    atoms = Atoms(
        cell=n_st.cell,
        numbers=n_st.numbers,
        positions=pos,
    )
    return atoms


def get_perms(
    atoms: Atoms,
    cyclic_group_ops: List[SymmOp],
    point_group_ops: List[SymmOp],
    symprec: float = 1e-2,
) -> Tuple[np.ndarray, List[SymmOp]]:
    """get the permutation table from symmetry operations

    Args:
        atoms:
        cyclic_group_ops:
        point_group_ops:
        symprec:

    Returns: permutation table
             rotation matrix (RM) = RM from point group @ RM from cyclic group
    """
    combs = list(itertools.product(point_group_ops, cyclic_group_ops))
    coords_car = atoms.positions
    coords_scaled = atoms.get_scaled_positions()
    coords_car_center = (
        atoms.get_scaled_positions() - [0.5, 0.5, 0.5]
    ) @ atoms.cell

    perms, rotation_matrix, translation_vector = [], [], []
    for ii, op in enumerate(combs):
        tmp_perm = np.ones((1, len(atoms.numbers)))[0]
        op1, op2 = op
        rotation_matrix.append(op1.rotation_matrix @ op2.rotation_matrix)
        translation_vector.append(
            op1.translation_vector + op2.translation_vector
        )

        for jj, site in enumerate(atoms):
            pos = (site.scaled_position - [0.5, 0.5, 0.5]) @ atoms.cell

            tmp = op1.operate(pos)
            idx1 = find_in_coord_list(coords_car_center, tmp, symprec)

            tmp1 = op2.operate(coords_car[idx1.item()])
            tmp1 = np.remainder(tmp1 @ np.linalg.inv(atoms.cell), [1, 1, 1])
            idx2 = find_in_coord_list(coords_scaled, tmp1, symprec)

            if idx2.size == 0:
                logging.ERROR("tolerance exceed while calculate perms")
            tmp_perm[jj] = idx2.item()
        perms.append(tmp_perm)
    perms_table, itp = np.unique(perms, axis=0, return_index=True)
    perms_table = perms_table.astype(np.int32)

    rotation_matrix = np.array(rotation_matrix)[itp]
    translation_vector = np.array(translation_vector)[itp]
    sym_operations = [
        SymmOp.from_rotation_and_translation(
            rotation_matrix[ii], translation_vector[ii]
        )
        for ii in range(len(itp))
    ]
    return perms_table, sym_operations


def get_perms_from_ops(
    atoms: Atoms,
    ops_sym: List[SymmOp],
    symprec: float = 1e-2,
    round: int = 4,
) -> np.ndarray:
    """get the permutation table from symmetry operations

    Args:
        atoms:
        symprec:

    Returns: permutation table
    """
    invcell = np.linalg.inv(atoms.cell)
    coords_center = atoms.positions - atoms.get_center_of_mass()
    coords_scaled_center = coords_center @ invcell
    coords_scaled_center[coords_center @ invcell >= 0.5] -= 1
    coords_scaled_center[coords_center @ invcell <= -0.5] += 1
    coords_center = coords_scaled_center @ atoms.cell

    perms = []
    for ii, op in enumerate(ops_sym):
        perms.append([])
        operated = op.operate_multi(coords_center)
        operated_scaled = operated @ invcell
        # much safer than remainder - no floating point issues
        operated_scaled[operated @ invcell >= 0.5] -= 1
        operated_scaled[operated @ invcell <= -0.5] += 1
        operated = operated_scaled @ atoms.cell
        for aid in range(len(coords_center)):
            equivalents = np.where(
                np.linalg.norm(operated[aid] - coords_center, axis=1) < symprec
            )[0]
            if len(equivalents) == 1:
                for eq in equivalents:
                    perms[-1].append(eq)
            else:
                print(ii, aid)
                raise ValueError
    perms_table = np.array(perms).astype(np.int32)
    return perms_table


def get_matrices(
    atoms: Atoms,
    ops_sym: List[SymmOp],
    symprec: float = 1e-5,
) -> List[np.ndarray]:
    """Build 3N x 3N representation matrices from
    symmetry operations and permutations.

    Args:
        atoms: ASE Atoms object.
        ops_sym: list of SymmOp symmetry operations.
        symprec: tolerance for permutation detection.

    Returns:
        list of np.ndarray: representation matrices of shape (3N, 3N).
    """
    perms_table = get_perms_from_ops(atoms, ops_sym, symprec=symprec)

    natoms = len(atoms.numbers)
    matrices = []
    for ii, perm in enumerate(perms_table):
        matrix = np.zeros((3 * natoms, 3 * natoms))
        for jj in range(natoms):
            idx = perm[jj]
            matrix[3 * idx : 3 * (idx + 1), 3 * jj : 3 * (jj + 1)] = ops_sym[
                ii
            ].rotation_matrix.copy()
        matrices.append(matrix)
    return matrices


def get_matrices_withPhase(
    atoms: Atoms,
    ops_sym: List[SymmOp],
    qpoint: float,
    symprec: float = 1e-3,
) -> List[np.ndarray]:
    """Build complex 3N x 3N representation matrices
    with Bloch phase factors.

    Args:
        atoms: ASE Atoms object.
        ops_sym: list of SymmOp symmetry operations.
        qpoint: q-point value along z for the phase factor exp(i*q*f).
        symprec: tolerance for permutation detection.

    Returns:
        list of complex np.ndarray of shape (3N, 3N).
    """
    perms_table = get_perms_from_ops(atoms, ops_sym, symprec=symprec)

    natoms = len(atoms.numbers)
    matrices = []
    for ii, perm in enumerate(perms_table):
        matrix = np.zeros((3 * natoms, 3 * natoms)).astype(np.complex128)
        phasefacter = ops_sym[ii].translation_vector[2]

        for jj in range(natoms):
            idx = perm[jj]
            matrix[3 * idx : 3 * (idx + 1), 3 * jj : 3 * (jj + 1)] = ops_sym[
                ii
            ].rotation_matrix.copy() * np.exp(1j * qpoint * phasefacter)
        matrices.append(matrix)
    return matrices


def affine_matrix_op(
    af1: np.ndarray, af2: np.ndarray, symprec: float = 1e-8
) -> np.ndarray:
    """Definition of group multiplication

    Args:
        af1: group element 1
        af2: group element 2

    Returns:

    """
    af1 = np.round(af1 / symprec).astype(int) * symprec
    af2 = np.round(af2 / symprec).astype(int) * symprec

    ro = af2[:3, :3] @ af1[:3, :3]
    tran = np.remainder(af2[:3, 3] + af2[:3, :3] @ af1[:3, 3], [1, 1, 1])

    af = np.eye(4)
    af[:3, :3] = ro
    af[:3, 3] = tran

    return af


def dimino_affine_matrix_and_character(
    generators: np.ndarray,
    character: np.ndarray,
    symec: float = 0.001,
) -> Tuple[np.ndarray, np.ndarray]:
    """

    Args:
        generators: the generators of point group
        symec: system precision

    Returns: all the group elements and correspond character

    """
    e_in = np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]])

    if character[0].ndim == 0:
        L_chara = np.array([np.complex128(1)])
        C_chara = np.array([np.complex128(1)])
    else:
        L_chara = np.array([np.complex128(1) * np.eye(character[0].shape[0])])
        C_chara = np.array([np.complex128(1) * np.eye(character[0].shape[0])])

    G = generators
    g, g1 = G[0].copy(), G[0].copy()
    g_chara, g1_chara = character[0].copy(), character[0].copy()
    L = np.array([e_in])
    while not ((g - e_in) < symec).all():
        L = np.vstack((L, [g]))
        L_chara = np.vstack((L_chara, [g_chara]))

        g = affine_matrix_op(g, g1)
        g_chara = np.dot(g_chara, g1_chara)

    for ii in range(len(G)):
        C = np.array([e_in])
        L1 = L.copy()
        L1_chara = L_chara.copy()

        more = True
        while more:
            more = False
            for jj, g in enumerate(list(C)):
                g_chara = C_chara[jj]

                for kk, ss in enumerate(G[: ii + 1]):
                    ss_chara = character[kk]
                    sg = affine_matrix_op(ss, g)
                    sg_chara = np.dot(ss_chara, g_chara)

                    itp = (abs((sg - L).sum(axis=1).sum(axis=1)) < 0.001).any()
                    if not itp:
                        if C.ndim == 3:
                            C = np.vstack((C, [sg]))
                            C_chara = np.vstack((C_chara, [sg_chara]))
                        else:
                            C = np.array((C, sg))
                            C_chara = np.array((C_chara, sg_chara))

                        if L.ndim == 3:
                            L = np.vstack(
                                (
                                    L,
                                    np.array(
                                        [affine_matrix_op(sg, t) for t in L1]
                                    ),
                                )
                            )
                            tmp = np.array(
                                [
                                    np.dot(sg_chara, t_chara)
                                    for t_chara in L1_chara
                                ]
                            )
                            L_chara = np.vstack((L_chara, tmp))
                        else:
                            L = np.array(
                                L,
                                np.array(
                                    [affine_matrix_op(sg, t) for t in L1]
                                ),
                            )
                        more = True
    L_chara_trace = []
    for lc in L_chara:
        if lc.ndim == 1:
            L_chara_trace.append(lc)
        else:
            L_chara_trace.append(np.trace(lc))
    return L, np.array(L_chara_trace)


def brute_force_generate_group(generators: np.ndarray, symec: float = 0.01):
    """Generate all group elements by brute-force
    multiplication of generators.

    Args:
        generators: array of 4x4 affine matrices as group generators.
        symec: tolerance for comparing matrix elements.

    Returns:
        np.ndarray: array of all group element affine matrices.
    """
    e_in = np.eye(4)
    G = generators
    L = np.array([e_in])
    while True:
        numL_old = len(L)
        for g in L:
            for h in G:
                gh = affine_matrix_op(g, h)
                judge = (abs(L - gh) < symec).all(axis=1).all(axis=1).any()
                if not judge:
                    L = np.concatenate((L, [gh]), axis=0)
        numL_new = len(L)
        if numL_old == numL_new:
            break
    return L


def brute_force_generate_group_subsequent(
    generators: np.ndarray, symec: float = 0.01
) -> Tuple[np.ndarray, List[List[int]]]:
    """generate all the group elements by brute force algorithm

    Args:
        generators: The most basic elements used to generate complete groups
        symec: tolerance

    Returns:
        L: all group elements
        L_seq: the multiplication order of the generators
    """

    e_in = np.eye(4)
    G = generators
    L = np.array([e_in])
    L_seq = [[0]]
    while True:
        numL_old = len(L)
        for ii, g in enumerate(L):
            tmp_seq1 = L_seq[ii]
            for jj, h in enumerate(G):
                gh = affine_matrix_op(g, h, symprec=1e-5)
                tmp_seq2 = tmp_seq1 + [jj + 1]
                judge = (abs(L - gh) < symec).all(axis=1).all(axis=1).any()
                if not judge:
                    L = np.concatenate((L, [gh]), axis=0)
                    L_seq.append(tmp_seq2)
        numL_new = len(L)
        if numL_old == numL_new:
            break
    return L, L_seq


def dimino(generators: np.ndarray, symec: int = 4) -> np.ndarray:
    """
    Generate all group elements from a set of generators
    using Dimino's algorithm.

    Dimino's algorithm performs coset enumeration: starting
    from the cyclic group generated by the first generator,
    it iteratively adds new cosets for each subsequent
    generator until the group is closed.

    Reference:
        Dimino, A. (1971). An algorithm for generating
        all group elements from a set of generators.
        See also: computational group theory texts.

    Args:
        generators: Array of shape (n_gen, 3, 3) containing
            the generator matrices of the point group. Each
            matrix is a 3x3 real orthogonal matrix
            representing a symmetry operation.
        symec: Number of decimal places for rounding when
            comparing matrices (to handle floating-point
            near-equalities). Default is 4.

    Returns:
        np.ndarray of shape (|G|, 3, 3): all group elements (symmetry
        operation matrices), including the identity element E.

    Example:
        >>> from pulgon_tools.utils import Cn, sigmaV, e
        >>> gens = np.array([Cn(6), sigmaV()])
        >>> all_ops = dimino(gens, symec=4)
        >>> all_ops.shape  # C6v has 12 elements
        (12, 3, 3)
    """
    G = generators

    # --- Step 1: Build the cyclic subgroup generated by G[0] ---
    # Start from E, repeatedly multiply by G[0] until we return to E.
    g = copy.deepcopy(G[0])
    g1 = copy.deepcopy(G[0])
    L = np.array(
        [e()]
    )  # L holds all known group elements; initialise with identity

    while not (np.round(g, symec) == e()).all():
        L = np.vstack((L, [g]))
        g = np.dot(g, g1)

    L = np.round(L, symec)

    # --- Step 2: Extend L by coset enumeration over each generator ---
    # For each generator G[i], try to find elements s*g
    # (s in G[0..i], g in coset C) that are not yet in L,
    # and if so, add the entire left-coset of that element.
    for ii in range(len(G)):
        C = np.array(
            [e()]
        )  # Current coset representatives; start with identity
        L1 = copy.deepcopy(
            L
        )  # Snapshot of L at the start of this generator's pass
        more = True

        while more:
            more = False
            for g in C:  # iterate over current coset representatives
                for s in G[: ii + 1]:  # multiply by all generators seen so far
                    sg = np.round(np.dot(s, g), symec)

                    # Check whether sg is already in L
                    already_in_L = (
                        (np.round(sg, symec) == L)
                        .all(axis=1)
                        .all(axis=1)
                        .any()
                    )

                    if not already_in_L:
                        # sg is a new coset representative
                        # Add it to C
                        C = np.vstack((C, [sg]))

                        # Add the entire left-coset  sg * L1  to L
                        new_coset = np.array([np.dot(sg, t) for t in L1])
                        L = np.vstack((L, new_coset))

                        more = True  # L grew, so we need another pass

    return L


def dimino_affine_matrix(
    generators: np.ndarray, symec: float = 0.01
) -> np.ndarray:
    """

    Args:
        generators: the generators of point group
        symec: system precision

    Returns: all the group elements and correspond character

    """
    e_in = np.eye(4)

    G = generators
    g, g1 = G[0].copy(), G[0].copy()
    L = np.array([e_in])
    while not ((g - e_in) < symec).all():
        L = np.vstack((L, [g]))
        g = affine_matrix_op(g, g1)
    for ii in range(len(G)):
        C = np.array([e_in])
        L1 = L.copy()
        more = True
        while more:
            more = False
            for g in list(C):
                for ss in G[: ii + 1]:
                    sg = affine_matrix_op(ss, g)
                    itp = (abs((sg - L).sum(axis=1).sum(axis=1)) < symec).any()
                    if not itp:
                        if C.ndim == 3:
                            C = np.vstack((C, [sg]))
                        if L.ndim == 3:
                            L = np.vstack(
                                (
                                    L,
                                    np.array(
                                        [affine_matrix_op(sg, t) for t in L1]
                                    ),
                                )
                            )
                        more = True
    L = np.unique(L, axis=0)
    return L


def dimino_affine_matrix_and_subsequent(
    generators: np.ndarray, symec: float = 0.001
) -> Tuple[np.ndarray, List[List[int]]]:
    """

    Args:
        generators: the generators of point group
        symec: system precision

    Returns: all the group elements and correspond character

    """
    e_in = np.eye(4)

    G = generators
    g, g1 = G[0].copy(), G[0].copy()
    g_subs, g1_subs = [1], [1]
    L = np.array([e_in])
    L_subs = [[0]]

    while not ((g - e_in) < symec).all():
        L = np.vstack((L, [g]))
        L_subs.append(g_subs.copy())

        g = affine_matrix_op(g, g1)
        g_subs = g_subs + g1_subs
    for ii in range(len(G)):
        C = np.array([e_in])
        C_subs = [[0]]

        L1 = L.copy()
        L1_subs = L_subs.copy()
        more = True
        while more:
            more = False
            for jj, g in enumerate(list(C)):
                g_subs = C_subs[jj]
                for kk, ss in enumerate(G[: ii + 1]):
                    ss_subs = [kk + 1]

                    sg = affine_matrix_op(ss, g)
                    sg_subs = ss_subs + g_subs

                    itp = (abs((sg - L).sum(axis=1).sum(axis=1)) < 0.001).any()
                    if not itp:
                        if C.ndim == 3:
                            C = np.vstack((C, [sg]))
                            C_subs.append(sg_subs)
                        else:
                            C = np.array((C, sg))
                            C_subs = C_subs.append(sg_subs)
                        if L.ndim == 3:
                            L = np.vstack(
                                (
                                    L,
                                    np.array(
                                        [affine_matrix_op(sg, t) for t in L1]
                                    ),
                                )
                            )
                            tmp = [sg_subs + t_subs for t_subs in L1_subs]
                            L_subs = L_subs + tmp

                        else:
                            L = np.array(
                                L,
                                np.array(
                                    [affine_matrix_op(sg, t) for t in L1]
                                ),
                            )
                            tmp = [sg_subs + t_subs for t_subs in L1_subs]
                            L_subs = L_subs + tmp
                        more = True
    return L, L_subs


def get_character(
    DictParams: dict, symprec: float = 1e-8
) -> Tuple[list, list, list]:
    """Compute symbolic representation matrices for the line group.

    Args:
        DictParams: dict with keys 'qpoints', 'nrot', 'order', 'family', 'a'.
        symprec: tolerance for symbolic evaluation.

    Returns:
        tuple of (characters, paras_values, paras_symbols).
    """
    characters, paras_values, paras_symbols = line_group_sympy(
        DictParams, symprec
    )
    return characters, paras_values, paras_symbols


def get_character_withparities(
    DictParams: dict, symprec: float = 1e-8
) -> Tuple[List[Any], List[Any], List[Any]]:
    """Compute symbolic representation matrices with parity operations.

    Args:
        DictParams: dict with keys 'qpoints', 'nrot', 'order', 'family', 'a'.
        symprec: tolerance for symbolic evaluation.

    Returns:
        tuple of (characters, paras_values, paras_symbols).
    """
    characters, paras_values, paras_symbols = line_group_sympy_withparities(
        DictParams, symprec
    )
    return characters, paras_values, paras_symbols


def get_character_num(
    DictParams: dict, symprec: float = 1e-8
) -> Tuple[np.ndarray, list, list]:
    """Compute numerical character table from representation matrices.

    Takes the trace of each representation matrix to obtain characters.

    Args:
        DictParams: dict with keys 'qpoints', 'nrot', 'order', 'family', 'a'.
        symprec: tolerance for symbolic evaluation.

    Returns:
        tuple of (characters, paras_values, paras_symbols) where
        characters is a numpy array of traces.
    """
    representation_mat, paras_values, paras_symbols = line_group_sympy(
        DictParams, symprec
    )

    characters = []
    for ii, rep_mat in enumerate(representation_mat):  # loop IR
        if rep_mat.ndim == 1:
            characters.append(rep_mat)
        else:
            characters.append(np.trace(rep_mat, axis1=1, axis2=2))
    characters = np.array(characters)
    return characters, paras_values, paras_symbols


def get_character_num_withparities(
    DictParams: dict, symprec: float = 1e-8
) -> Tuple[np.ndarray, list, list]:
    """Compute numerical character table with parity operations.

    Takes the trace of each representation matrix (including parities)
    to obtain characters.

    Args:
        DictParams: dict with keys 'qpoints', 'nrot', 'order', 'family', 'a'.
        symprec: tolerance for symbolic evaluation.

    Returns:
        tuple of (characters, paras_values, paras_symbols) where
        characters is a numpy array of traces.
    """
    (
        representation_mat,
        paras_values,
        paras_symbols,
    ) = line_group_sympy_withparities(DictParams, symprec)
    characters = []
    for ii, rep_mat in enumerate(representation_mat):  # loop IR
        if rep_mat.ndim == 1:
            characters.append(rep_mat)
        else:
            characters.append(np.trace(rep_mat, axis1=1, axis2=2))
    characters = np.array(characters)
    return characters, paras_values, paras_symbols


def get_sym_constrains_matrices_M(
    ops: np.ndarray,
    permutations: np.ndarray,
    diminsion: int = 3,
) -> ss.spmatrix:
    """M K = 0

    :param ops:
    :param permutations:
    :param diminsion: spatial dimension (default 3)
    :return:

    """

    if permutations.ndim == 2:
        natom = len(permutations[0])
    elif permutations.ndim == 1:
        natom = len(permutations)
        permutations = np.array([permutations])
    else:
        logging.ERROR("error for permutations' ndim")

    if len(ops.shape) == 2:
        ops = np.array([ops])

    size1 = diminsion**2
    I = np.eye(size1)
    M = []

    idx1 = np.repeat(np.arange(natom), natom)
    idx2 = np.tile(np.arange(natom), natom)

    tmp1 = (idx1 * natom + idx2) * size1
    tmp2 = (idx1 * natom + idx2 + 1) * size1
    tmp3 = np.linspace(tmp1, tmp2, size1 + 1).astype(np.int64)[:-1, :].T
    for ii, op in enumerate(ops):
        print("now run in %s operation" % ii)
        perm = permutations[ii]
        C = np.einsum(
            "ij,kl->ikjl",
            op[:diminsion, :diminsion],
            op[:diminsion, :diminsion],
        ).reshape(size1, size1)
        x = ss.csc_matrix((size1 * natom**2, size1 * natom**2))
        if (perm == np.arange(natom)).all():
            M.append(x)
            continue

        pidx1 = perm[idx1]
        pidx2 = perm[idx2]

        ptmp1 = (pidx1 * natom + pidx2) * size1
        ptmp2 = (pidx1 * natom + pidx2 + 1) * size1
        ptmp3 = np.linspace(ptmp1, ptmp2, size1 + 1).astype(np.int64)[:-1, :].T

        itp1 = np.repeat(tmp3, size1, axis=1)
        itp2 = np.tile(tmp3, (1, size1))
        pitp2 = np.tile(ptmp3, (1, size1))
        xl = x.tolil()
        xl[itp1, itp2] = C.flatten()
        xl[itp1, pitp2] = -I.flatten()  #
        M.append(xl)

    M = scipy.sparse.vstack((M))
    return M


def _calc_dists(
    atoms: Atoms, tolerance: float = 1e-3
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Return the distances between atoms in the supercell, their
    degeneracies and the associated displacements along OZ.
    """
    MIN_DELTA = -1
    MAX_DELTA = 1
    positions = atoms.positions
    cell = atoms.cell
    n_satoms = positions.shape[0]
    d2s = np.empty((MAX_DELTA - MIN_DELTA + 1, n_satoms, n_satoms))

    # TODO: This could not be enough and eventually we
    # should do a proper check for the shortest distance.
    for j, j_c in enumerate(range(MIN_DELTA, MAX_DELTA + 1)):
        shifted_positions = positions + (j_c * cell[2, :])[np.newaxis, :]
        d2s[j, :, :] = sp.spatial.distance.cdist(
            positions, shifted_positions, "sqeuclidean"
        )
        # d2s[j, :, :] = sp.spatial.distance.cdist(
        #     positions, shifted_positions)
    d2min = d2s.min(axis=0)
    dmin = np.sqrt(d2min)
    degenerate = np.abs(d2s - d2min) < tolerance
    nequi = degenerate.sum(axis=0, dtype=int)

    maxequi = nequi.max()
    shifts = np.empty((n_satoms, n_satoms, maxequi))
    sorting = np.argsort(np.logical_not(degenerate), axis=0)
    shifts = np.transpose(sorting[:maxequi, :, :], (1, 2, 0)).astype(np.intc)
    shifts = np.asarray(range(MIN_DELTA, MAX_DELTA + 1))[shifts]
    return (dmin, nequi, shifts)


def get_continum_constrains_matrices_M_for_conpact_fc(
    phonon,
) -> ss.coo_array:
    """Build constraint matrix enforcing acoustic and rotational sum rules.

    Constructs a sparse matrix M such that M @ K = 0 encodes translational
    (acoustic), Born-Huang, and Huang invariance conditions on compact
    force constants.

    Args:
        phonon: Phonopy object with force_constants and supercell set.

    Returns:
        scipy sparse COO matrix encoding all continuum constraints.
    """
    IFC = phonon.force_constants.copy()
    scell = phonon.supercell

    symbols = scell.symbols
    cell = scell.cell
    positions = (
        (scell.scaled_positions + np.asarray([0.5, 0.5, 0.0])[np.newaxis, :])
        % 1.0
    ) @ cell
    ase_atoms = ase.Atoms(symbols, positions, cell=cell, pbc=True)
    dists, degeneracy, shifts = _calc_dists(ase_atoms)
    n_satoms = len(symbols)

    # %%
    average_delta = np.zeros((n_satoms, n_satoms, 3))
    for i in range(n_satoms):
        for j in range(n_satoms):
            n_elements = degeneracy[i, j]
            for i_d in range(n_elements):
                average_delta[i, j, :] += (
                    positions[j, :]
                    - positions[i, :]
                    + shifts[i, j, i_d] * cell[2, :]
                )
            average_delta[i, j, :] /= n_elements

    average_pos = scell.scaled_positions @ cell

    # %%
    average_products = np.zeros((n_satoms, n_satoms, 3, 3))
    for i in range(n_satoms):
        for j in range(n_satoms):
            n_elements = degeneracy[i, j]
            for i_d in range(n_elements):
                delta = (
                    positions[j, :]
                    - positions[i, :]
                    + shifts[i, j, i_d] * cell[2, :]
                )
                average_products[i, j, :, :] += np.outer(delta, delta)
            average_products[i, j, :, :] /= n_elements

    M1 = ss.coo_array(([]))

    # Append the acoustic sum rules originating from translations.
    n_atoms, n_satoms = IFC.shape[:2]
    n_rows = M1.shape[0]
    rows = M1.row.tolist()
    cols = M1.col.tolist()
    data = M1.data.tolist()

    # translational sum rules
    for i in range(n_atoms):
        for alpha in range(3):
            for beta in range(3):
                for j in range(n_satoms):
                    rows.append(n_rows)
                    cols.append(
                        np.ravel_multi_index((i, j, alpha, beta), IFC.shape)
                    )
                    data.append(1.0)
                n_rows += 1

    # The same but for rotations (Born-Huang).
    positions = phonon.supercell.positions
    for i in range(n_atoms):
        for alpha in range(3):
            for beta in range(3):
                for gamma in range(3):
                    for j in range(n_satoms):
                        r_ij = average_delta[phonon.primitive.p2s_map[i], j]
                        rows.append(n_rows)
                        cols.append(
                            np.ravel_multi_index(
                                (i, j, alpha, beta), IFC.shape
                            )
                        )
                        data.append(r_ij[gamma])
                        rows.append(n_rows)
                        cols.append(
                            np.ravel_multi_index(
                                (i, j, alpha, gamma), IFC.shape
                            )
                        )
                        data.append(-r_ij[beta])
                    n_rows += 1

    # And fthe Huang invariances, also for rotation.
    for alpha in range(3):
        for beta in range(3):
            for gamma in range(3):
                for delta in range(3):
                    for i in range(n_atoms):
                        for j in range(n_satoms):
                            products = average_products[
                                phonon.primitive.p2s_map[i], j
                            ]
                            rows.append(n_rows)
                            cols.append(
                                np.ravel_multi_index(
                                    (i, j, alpha, beta), IFC.shape
                                )
                            )
                            data.append(products[gamma, delta])
                            rows.append(n_rows)
                            cols.append(
                                np.ravel_multi_index(
                                    (i, j, gamma, delta), IFC.shape
                                )
                            )
                            data.append(-products[alpha, beta])
                    n_rows += 1

    # Make sure the IFC matrix is symmetric.
    for i in range(n_atoms):
        for j in range(n_atoms):
            for alpha in range(3):
                for beta in range(3):
                    rows.append(n_rows)
                    cols.append(
                        np.ravel_multi_index(
                            (i, phonon.primitive.p2s_map[j], alpha, beta),
                            IFC.shape,
                        )
                    )

                    data.append(1.0)
                    rows.append(n_rows)
                    cols.append(
                        np.ravel_multi_index(
                            (j, phonon.primitive.p2s_map[i], beta, alpha),
                            IFC.shape,
                        )
                    )
                    data.append(-1.0)
                    n_rows += 1

    # Rebuild the sparse matrix.
    M1 = ss.coo_array((data, (rows, cols)), shape=(n_rows, IFC.size))
    return M1


def get_IFCSYM_from_cvxpy_M(M: ss.spmatrix, IFC: np.ndarray) -> np.ndarray:
    """Symmetrize force constants by solving a constrained optimization.

    Minimizes ||x - IFC||^2 subject to M @ x = 0 using CVXPY.

    Args:
        M: sparse constraint matrix encoding symmetry/sum rules.
        IFC: original force constant array.

    Returns:
        np.ndarray: symmetrized force constants with the same shape as IFC.
    """
    flat_IFCs = IFC.ravel()
    x = cp.Variable(IFC.size)
    cost = cp.sum_squares(x - flat_IFCs)
    prob = cp.Problem(cp.Minimize(cost), [M @ x == 0])
    prob.solve()
    IFC_sym = x.value.reshape(IFC.shape)
    return IFC_sym


def get_symbols_from_ops(
    ops_sym: List[np.ndarray],
) -> List[str]:
    """
    Get the symbols of point group operations from the rotation matrix.

    Args:
        ops_sym: A list of rotation matrix of symmetry operations.

    Returns:
        symbols: A list of point group symbols.

    Notes:
        The point group operations are as follows:
        E: Identity
        sigmaH: Reflection in the horizontal mirror plane.
        sigmaV: Reflection in the vertical mirror plane.
        U: Reflection in the plane perpendicular to the axis.
        Cn: Rotation by 2 * pi / n about the axis.
        Sn: Rotation by 2 * pi / n about the axis
            followed by reflection in the plane
            perpendicular to the axis.
    """
    symbols = []
    for op in ops_sym:
        op_ro = op[:3, :3]

        val1, _ = np.linalg.eig(op_ro)
        itp1 = np.isclose(val1, 1, atol=1e-4)
        itp2 = np.isclose(val1, -1, atol=1e-4)

        if np.logical_or(itp1, itp2).all():  # test if it is a reflection h,v,U
            if itp1.sum() == 3:
                symbols.append("E")
            elif itp1.sum() == 2 and itp2.sum() == 1:
                tmp = np.where(itp2)[0].item()
                if tmp == 2:
                    symbols.append("sigmaH")
                else:
                    symbols.append("sigmaV")
            elif itp1.sum() == 1 and itp2.sum() == 2 and itp2[2]:
                symbols.append("U")
            else:
                logging.ERROR("The Eigenvalues are very strange")
        else:  # test if it is a rotation Cn
            tmp = 1 / (-np.log(val1) / 2 / np.pi * 1j)[:2]

            if np.isclose(
                tmp[0].real, abs(tmp[1].real), atol=1e-4
            ) and np.isclose(val1[2].real, 1, atol=1e-4):
                num = int(np.round(tmp[0].real))
                symbols.append("C%d" % num)
            elif np.isclose(
                tmp[0].real, abs(tmp[1].real), atol=1e-4
            ) and np.isclose(val1[2].real, -1, atol=1e-4):
                num = int(np.round(tmp[0].real))
                symbols.append("S%d" % num)
            else:
                logging.ERROR("The Eigenvalues are very strange")
    return symbols


def divide_irreps(
    vec: np.ndarray,
    adapted: np.ndarray,
    dimensions: List[int],
) -> np.ndarray:
    """
    Project vectors into IR space according to the adapted basis.

    Parameters
    ----------
    vec : array of shape (n,m), where n is the number
        of vectors and m is the dimension
        The vector to be divided
    adapted : array of shape (m,m). The basis vectors are arranged in columns.
        The adapted basis
    dimensions : list of int. Sum(dimensions) = m.
        The dimension of each irrep

    Returns
    -------
    means : array of shape (n,k), where k is the number of irreps
        The projected length of each irrep
    """
    tmp1 = vec @ adapted.conj()
    start = 0
    means, vectors = [], []
    for im, dim in enumerate(dimensions):
        end = start + dim
        if vec.ndim == 1:
            means.append((np.abs(tmp1[start:end]) ** 2).sum())
        else:
            means.append((np.abs(tmp1[:, start:end]) ** 2).sum(axis=1))
        start = copy.copy(end)
    means = np.array(means)
    if means.ndim > 1:
        means = means.T
    return np.array(means)


def angle_between_points(
    A: Union[list, np.ndarray],
    B: Union[list, np.ndarray],
    C: Union[list, np.ndarray],
) -> float:
    """
    Calculate angle between three points. Point b is the vertex

    Parameters
    ----------
    A : The first point coordinates
    B : The second point coordinates
    C : The third point coordinates

    Returns
    -------
    angle : float
        The angle between points A, B and C in degrees.
    """
    BA = np.array(A) - np.array(B)
    BC = np.array(C) - np.array(B)

    cosine_angle = np.dot(BA, BC) / (np.linalg.norm(BA) * np.linalg.norm(BC))
    cosine_angle = np.clip(cosine_angle, -1.0, 1.0)
    angle_rad = np.arccos(cosine_angle)
    return np.degrees(angle_rad)



# Copyright 2023-2026 The PULGON Project Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied. See the License for the specific language governing
# permissions and limitations under the License.

import argparse

import numpy as np
import phonopy
import pytest
from ase import Atoms
from phonopy.file_IO import parse_FORCE_CONSTANTS

from pulgon_tools.force_constant_correction import (
    build_constraint_matrix,
    calc_dists,
    parse_bool_list,
    parse_int_list,
    solve_fcs,
    str2list,
)

pytest_plugins = ["pytest-datadir"]

FCS_DATA = "test/data/fcs"


@pytest.fixture
def phonon_obj():
    """Load the phonopy object from test data."""
    return phonopy.load(
        phonopy_yaml=f"{FCS_DATA}/phonopy.yaml",
        force_constants_filename=f"{FCS_DATA}/FORCE_CONSTANTS",
    )


@pytest.fixture
def supercell_atoms(phonon_obj):
    """Build an ASE Atoms object from the phonopy supercell."""
    scell = phonon_obj.supercell
    positions = (scell.scaled_positions % 1.0) @ scell.cell
    return Atoms(scell.symbols, positions, cell=scell.cell)


class TestCalcDists:
    """Test distance calculation with periodic images."""

    def test_dists_output(self, supercell_atoms):
        dists, degeneracy, shifts = calc_dists(supercell_atoms)
        n = len(supercell_atoms)
        min_d = dists[dists > 0.01].min()
        assert dists.shape == (n, n)
        assert degeneracy.shape == (n, n)
        assert shifts.shape[0] == n and shifts.shape[1] == n
        assert np.allclose(np.diag(dists), 0.0)
        assert 1.0 < min_d < 5.0
        assert (degeneracy >= 1).all()
        assert np.allclose(dists, dists.T, atol=1e-10)


class TestBuildConstraintMatrix:
    """Test constraint matrix construction."""

    def test_output_types(self, phonon_obj):
        M, IFC = build_constraint_matrix(phonon_obj)
        assert hasattr(M, "nnz")
        assert isinstance(IFC, np.ndarray)
        n_prim = len(phonon_obj.primitive)
        n_super = len(phonon_obj.supercell)
        assert IFC.shape == (n_prim, n_super, 3, 3)
        assert M.shape[1] == IFC.size
        assert M.nnz > 0

    def test_raw_ifc_violates_constraints(self, phonon_obj):
        """Raw IFC should have nonzero constraint violation."""
        M, IFC = build_constraint_matrix(phonon_obj)
        violation = np.abs(M.dot(IFC.ravel())).sum()
        assert violation > 1.0


class TestSolveFcs:
    """Test force constant correction solver."""

    @pytest.fixture
    def constraint_data(self, phonon_obj):
        M, IFC = build_constraint_matrix(phonon_obj)
        return M, IFC

    def test_convex_opt_satisfies_constraints(self, constraint_data):
        """After correction, M @ x should be nearly zero."""
        M, IFC = constraint_data
        IFC_sym = solve_fcs(IFC, M, methods="convex_opt")
        violation = np.abs(M.dot(IFC_sym.ravel())).sum()
        fcs_ref = parse_FORCE_CONSTANTS(
            f"{FCS_DATA}/FORCE_CONSTANTS_correction"
        )
        assert violation < 1e-6
        assert not np.allclose(IFC_sym, IFC)
        assert np.allclose(IFC_sym, fcs_ref, atol=0.15)

    def test_ridge_model_reduces_violation(self, constraint_data):
        """Ridge regression should reduce constraint violation."""
        M, IFC = constraint_data
        violation_before = np.abs(M.dot(IFC.ravel())).sum()
        IFC_sym = solve_fcs(IFC, M, methods="ridge_model")
        violation_after = np.abs(M.dot(IFC_sym.ravel())).sum()
        assert IFC_sym.shape == IFC.shape
        assert violation_after < violation_before


class TestParseIntList:
    """Test parse_int_list argument parser."""

    def test_valid_input(self):
        assert parse_int_list("[5, 5, 1]") == [5, 5, 1]

    def test_wrong_length(self):
        with pytest.raises(argparse.ArgumentTypeError):
            parse_int_list("[1, 2]")

    def test_non_int(self):
        with pytest.raises(argparse.ArgumentTypeError):
            parse_int_list("[1.0, 2.0, 3.0]")

    def test_invalid_string(self):
        with pytest.raises(argparse.ArgumentTypeError):
            parse_int_list("not_a_list")


class TestParseBoolList:
    """Test parse_bool_list argument parser."""

    def test_valid_input(self):
        assert parse_bool_list("[True, True, False]") == [True, True, False]

    def test_wrong_length(self):
        with pytest.raises(argparse.ArgumentTypeError):
            parse_bool_list("[True, False]")

    def test_non_bool(self):
        with pytest.raises(argparse.ArgumentTypeError):
            parse_bool_list("[1, 0, 1]")


class TestStr2List:
    """Test str2list argument parser."""

    def test_valid_input(self):
        result = str2list("[[0, 0, 0], [0.5, 0, 0]]")
        assert result == [[0, 0, 0], [0.5, 0, 0]]

    def test_none_input(self):
        assert str2list(None) is None

    def test_invalid_string(self):
        with pytest.raises(argparse.ArgumentTypeError):
            str2list("not valid python")



# Copyright 2023-2026 The PULGON Project Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied. See the License for the specific language governing
# permissions and limitations under the License.

import argparse
from typing import List, Tuple, Union

import numpy as np
import sympy
from ase import Atoms
from ase.io.vasp import write_vasp
from pulp import LpMinimize, LpProblem, LpVariable, value
from pymatgen.core.operations import SymmOp
from scipy.optimize import fsolve

from pulgon_tools.utils import Cn, brute_force_generate_group


def cyl2car(
    cyl: Union[list, np.ndarray],
) -> np.ndarray:
    """Convert cylindrical coordinates (phi, r, z) to Cartesian (x, y, z).

    Args:
        cyl: array-like of [phi, r, z] in cylindrical coordinates.

    Returns:
        np.ndarray of [x, y, z] in Cartesian coordinates.
    """
    car = np.array([cyl[1] * np.cos(cyl[0]), cyl[1] * np.sin(cyl[0]), cyl[2]])
    return car


def helical_group_analysis(
    a1: np.ndarray,
    a2: np.ndarray,
    n1: int,
    n2: int,
    L1: float,
) -> tuple:
    """Compute helical group parameters for a chiral nanotube (n1, n2).

    Derives the screw-axis order q, translational pitch f, tube radius r,
    helical quantum number R, chiral vector direction Ch, translational
    period t, translation indices (t1, t2), and GCD of (n1, n2).

    Args:
        a1: first 2D lattice vector of the hexagonal sheet.
        a2: second 2D lattice vector of the hexagonal sheet.
        n1: first chiral index.
        n2: second chiral index.
        L1: lattice constant (length of a1).

    Returns:
        symmetry indices: (q, f, r, R, Ch, t, t1, t2, n_gcd).
    """
    n_gcd = np.gcd(n1, n2)
    n1_tilde = int(n1 / n_gcd)
    n2_tilde = int(n2 / n_gcd)

    t_gcd = np.gcd(2 * n2_tilde + n1_tilde, 2 * n1_tilde + n2_tilde)
    t1 = -int((2 * n2_tilde + n1_tilde) / t_gcd)
    t2 = int((2 * n1_tilde + n2_tilde) / t_gcd)
    t = np.linalg.norm(t1 * a1 + t2 * a2)
    q_tilde = int(np.linalg.det([[n1_tilde, n2_tilde], [t1, t2]]))
    q = n_gcd * q_tilde

    f = t / q_tilde
    D = (
        L1
        / np.pi
        * n_gcd
        * np.sqrt(n1_tilde**2 + n1_tilde * n2_tilde + n2_tilde**2)
    )
    r = D / 2

    prob = LpProblem("IntegerProgrammingExample", LpMinimize)
    h1 = LpVariable("h1", lowBound=1, cat="Integer")
    h2 = LpVariable("h2", lowBound=1, cat="Integer")
    prob += h1 + h2
    prob += n1_tilde * h2 - n2_tilde * h1 == 1
    prob.solve()

    h1 = int(value(h1))
    h2 = int(value(h2))
    R = h1 * t2 - h2 * t1

    p_tilde = R ** (sympy.totient(q_tilde) - 1)
    p = n_gcd * p_tilde

    Ch = n1 * a1 + n2 * a2
    Ch = Ch / np.linalg.norm(Ch)
    return q, f, r, R, Ch, t, t1, t2, n_gcd


def bond_constraints_equations(
    variables: np.ndarray,
    pos_cyl1: np.ndarray,
    pos_cyl2: np.ndarray,
    pos_cyl3: np.ndarray,
    pos_cyl4: np.ndarray,
    bond_length: float,
) -> List[float]:
    """Nonlinear equations enforcing equal bond lengths on a nanotube surface.

    Solves for (del_phi, del_r, del_z) corrections to pos_cyl1 such that
    its distances to pos_cyl2, pos_cyl3, and pos_cyl4 are all equal to
    bond_length. Used as input to scipy.optimize.fsolve.

    Args:
        variables: array of [del_phi, del_r, del_z] corrections.
        pos_cyl1: cylindrical coordinates [phi, r, z] of the atom to adjust.
        pos_cyl2: cylindrical coordinates of the first neighbor.
        pos_cyl3: cylindrical coordinates of the second neighbor.
        pos_cyl4: cylindrical coordinates of the third neighbor.
        bond_length: target bond length.

    Returns:
        list of 3 residual equations [eq1, eq2, eq3].
    """
    del_phi, del_r, del_z = variables
    pos_car1 = np.array(
        [
            (pos_cyl1[1] + del_r) * np.cos(pos_cyl1[0] + del_phi),
            (pos_cyl1[1] + del_r) * np.sin(pos_cyl1[0] + del_phi),
            pos_cyl1[2] + del_z,
        ]
    )
    pos_car2 = np.array(
        [
            pos_cyl2[1] * np.cos(pos_cyl2[0]),
            pos_cyl2[1] * np.sin(pos_cyl2[0]),
            pos_cyl2[2],
        ]
    )
    pos_car3 = np.array(
        [
            pos_cyl3[1] * np.cos(pos_cyl3[0]),
            pos_cyl3[1] * np.sin(pos_cyl3[0]),
            pos_cyl3[2],
        ]
    )
    pos_car4 = np.array(
        [
            pos_cyl4[1] * np.cos(pos_cyl4[0]),
            pos_cyl4[1] * np.sin(pos_cyl4[0]),
            pos_cyl4[2],
        ]
    )
    eq1 = np.linalg.norm(pos_car1 - pos_car2) - np.linalg.norm(
        pos_car1 - pos_car3
    )
    eq2 = np.linalg.norm(pos_car1 - pos_car2) - np.linalg.norm(
        pos_car1 - pos_car4
    )
    eq3 = np.linalg.norm(pos_car1 - pos_car2) - bond_length
    return [eq1, eq2, eq3]


def generate_symcell_and_linegroup_elements(
    a1: np.ndarray,
    a2: np.ndarray,
    Ch: np.ndarray,
    t1: int,
    t2: int,
    r: float,
    bond_length: float,
    delta_Z: float,
    symbol1: Union[int, str] = 74,
    symbol2: Union[int, str] = 16,
    tol_round: int = 10,
) -> Tuple[np.ndarray, np.ndarray]:
    """Generate the symmetry cell (monomer) for a MoS2-type chiral nanotube.

    Maps flat-sheet atomic positions onto the tube surface in cylindrical
    coordinates, then optimizes chalcogen positions to satisfy bond-length
    constraints via fsolve.

    Args:
        a1: first 2D lattice vector.
        a2: second 2D lattice vector.
        Ch: unit chiral vector direction.
        t1: first translational index.
        t2: second translational index.
        r: tube radius.
        bond_length: target bond length between atoms.
        delta_Z: vertical offset of chalcogen atoms from the metal layer.
        symbol1: atomic number of the metal atom (default 74, W).
        symbol2: atomic number of the chalcogen atom (default 16, S).
        tol_round: decimal precision for rounding intermediate values.

    Returns:
        tuple of (pos_cyl, symbols) where pos_cyl is an (N, 3) array of
        cylindrical coordinates [phi, r, z] and symbols is an (N,) array
        of atomic numbers.
    """
    pos1 = 1 / 3 * a1 + 1 / 3 * a2
    pos2 = 2 / 3 * a1 + 2 / 3 * a2
    pos_auxiliary1 = 4 / 3 * a1 + 1 / 3 * a2
    pos_auxiliary2 = 1 / 3 * a1 + 4 / 3 * a2

    r1 = r + delta_Z
    r2 = r - delta_Z

    z1 = np.sqrt(
        np.round(np.linalg.norm(pos1) ** 2 - np.dot(pos1, Ch) ** 2, tol_round)
    )
    z2 = np.sqrt(
        np.round(np.linalg.norm(pos2) ** 2 - np.dot(pos2, Ch) ** 2, tol_round)
    )
    z_auxiliary1 = np.sign((t1 * a1 + t2 * a2) @ pos_auxiliary1) * np.sqrt(
        np.linalg.norm(pos_auxiliary1) ** 2 - np.dot(pos_auxiliary1, Ch) ** 2
    )
    z_auxiliary2 = np.sign((t1 * a1 + t2 * a2) @ pos_auxiliary2) * np.sqrt(
        np.linalg.norm(pos_auxiliary2) ** 2 - np.dot(pos_auxiliary2, Ch) ** 2
    )

    phi1 = np.sqrt(np.linalg.norm(pos1) ** 2 - z1**2) / r
    phi2 = np.sqrt(np.linalg.norm(pos2) ** 2 - z2**2) / r
    phi3 = np.sqrt(np.linalg.norm(pos2) ** 2 - z2**2) / r
    phi_auxiliary1 = (
        np.sqrt(np.linalg.norm(pos_auxiliary1) ** 2 - z_auxiliary1**2) / r
    )
    phi_auxiliary2 = (
        np.sqrt(np.linalg.norm(pos_auxiliary2) ** 2 - z_auxiliary2**2) / r
    )

    pos_cyl0 = [phi2, r1, z2]
    pos_cyl1 = [phi3, r2, z2]
    pos_cyl2 = [phi1, r, z1]

    pos_cyl3 = [phi_auxiliary1, r, z_auxiliary1]
    pos_cyl4 = [phi_auxiliary2, r, z_auxiliary2]

    initial_guess = np.array([0, 0, 0])
    solutions1 = fsolve(
        bond_constraints_equations,
        initial_guess,
        args=(pos_cyl0, pos_cyl2, pos_cyl3, pos_cyl4, bond_length),
    )
    solutions2 = fsolve(
        bond_constraints_equations,
        initial_guess,
        args=(pos_cyl1, pos_cyl2, pos_cyl3, pos_cyl4, bond_length),
    )

    pos_cyl = np.array(
        [pos_cyl2, pos_cyl0 + solutions1, pos_cyl1 + solutions2]
    )
    symbols = np.array([symbol1, symbol2, symbol2])
    return pos_cyl, symbols


def get_nanotube_from_n1n2(
    n1: int,
    n2: int,
    symbol1: Union[int, str],
    symbol2: Union[int, str],
    bond_length: float,
    delta_Z: float,
) -> Atoms:
    """Build a complete MoS2-type chiral nanotube from chiral indices (n1, n2).

    Computes helical group parameters, generates the symmetry cell, applies
    all line group operations, and returns a deduplicated ASE Atoms object.

    Args:
        n1: first chiral index.
        n2: second chiral index.
        symbol1: atomic number or symbol of the metal atom.
        symbol2: atomic number or symbol of the chalcogen atom.
        bond_length: metal-chalcogen bond length in Angstrom.
        delta_Z: vertical offset of chalcogen atoms from the metal layer.

    Returns:
        ase.Atoms: the nanotube structure with periodic boundary along z.
    """
    L1 = np.sqrt(bond_length**2 - delta_Z**2) * np.sqrt(
        2 - 2 * np.cos(2 * np.pi / 3)
    )  # the length of hex lattice
    a1 = L1 * np.array([1, 0])
    a2 = np.array([L1 * np.cos(np.pi / 3), L1 * np.sin(np.pi / 3)])
    q, f, r, R, Ch, t, t1, t2, n_gcd = helical_group_analysis(
        a1, a2, n1, n2, L1
    )

    coo_cyl, symbols = generate_symcell_and_linegroup_elements(
        a1,
        a2,
        Ch,
        t1,
        t2,
        r,
        bond_length=bond_length,
        delta_Z=delta_Z,
        symbol1=symbol1,
        symbol2=symbol2,
    )
    car_x = coo_cyl[:, 1] * np.cos(coo_cyl[:, 0])
    car_y = coo_cyl[:, 1] * np.sin(coo_cyl[:, 0])
    pos = np.hstack(
        (car_x[np.newaxis].T, car_y[np.newaxis].T, coo_cyl[:, 2][np.newaxis].T)
    )

    distance = np.max(pos[:, :2])
    cell = np.array(
        [[distance * 4.5, 0, 0], [0, distance * 4.5, 0], [0, 0, t]]
    )

    generators = [
        SymmOp.from_rotation_and_translation(
            rotation_matrix=Cn(n_gcd), translation_vec=[0, 0, 0]
        ).affine_matrix,
        SymmOp.from_rotation_and_translation(
            rotation_matrix=Cn(q / R), translation_vec=[0, 0, 1 / int(t / f)]
        ).affine_matrix,
    ]
    ops = brute_force_generate_group(generators, symec=1e-2)

    pos_car = np.array([])
    ops_sym = [
        SymmOp.from_rotation_and_translation(
            rotation_matrix=line[:3, :3], translation_vec=line[:3, 3] * t
        )
        for line in ops
    ]
    for op in ops_sym:
        for line in pos:
            tmp = op.operate(line)
            if pos_car.size == 0:
                pos_car = tmp[np.newaxis]
            else:
                judge = (
                    np.sqrt((pos_car - tmp) ** 2).sum(axis=1) < 1e-8
                ).any()
                if not judge:
                    pos_car = np.vstack((pos_car, tmp))
    symbols = np.tile(symbols, len(ops_sym))
    pos_car = np.array(pos_car)

    scaled = np.round(
        np.remainder(np.dot(pos_car, np.linalg.inv(cell)), [1, 1, 1]), 10
    )
    scaled, index = np.unique(scaled, axis=0, return_index=True)
    symbols = symbols[index]
    new_atom = Atoms(scaled_positions=scaled, cell=cell, symbols=symbols)
    return new_atom


def main():
    """CLI entry point for generating a chiral nanotube
    and saving as POSCAR."""
    parser = argparse.ArgumentParser(
        description="generating chiral nanotube structure from (n1, n2) indices"
    )

    parser.add_argument(
        "-c",
        "--chirality",
        default=(10, 10),
    )

    parser.add_argument(
        "-b",
        "--symbol",
        default=("Mo", "S"),
    )

    parser.add_argument("-l", "--bond_length", default=2.43, type=float)

    parser.add_argument("-d", "--delta_Z", default=1.57, type=float)

    parser.add_argument(
        "-s",
        "--st_name",
        type=str,
        default="POSCAR",
        help="the saved file name",
    )

    args = parser.parse_args()

    n1, n2 = eval(args.chirality)[0], eval(args.chirality)[1]

    symbol1, symbol2 = eval(args.symbol)[0], eval(args.symbol)[1]
    bond_length = args.bond_length
    delta_Z = args.delta_Z
    filename = args.st_name

    new_atom = get_nanotube_from_n1n2(
        n1, n2, symbol1, symbol2, bond_length, delta_Z
    )
    pos = (new_atom.get_scaled_positions() - [0.5, 0.5, 0]) % 1 @ new_atom.cell
    new_atom = Atoms(
        positions=pos, cell=new_atom.cell, numbers=new_atom.numbers
    )
    write_vasp(filename, new_atom, direct=True, sort=True)


if __name__ == "__main__":
    main()



# Copyright 2023-2026 The PULGON Project Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied. See the License for the specific language governing
# permissions and limitations under the License.


import argparse
from typing import List, Tuple, Union

import numpy as np
from ase import Atom, Atoms
from ase.io import read
from pymatgen.core.operations import SymmOp

from pulgon_tools.detect_generalized_translational_group import (
    CyclicGroupAnalyzer,
)
from pulgon_tools.detect_point_group import LineGroupAnalyzer
from pulgon_tools.line_group_table import get_family_num_from_sym_symbol
from pulgon_tools.utils import (
    brute_force_generate_group_subsequent,
    find_axis_center_of_nanotube,
    get_character_num_withparities,
    get_character_withparities,
)


def get_linegroup_symmetry_dataset(
    poscar: Union[str, Atom, Atoms],
) -> Tuple[Atoms, int, int, float, List[SymmOp], List[List[int]]]:
    """Extract the full line group symmetry dataset from a structure.

    Detects the axial point group and cyclic group, then generates all
    symmetry operations via brute-force group multiplication.

    Args:
        poscar: path to a POSCAR file, or an ASE Atom/Atoms object.

    Returns:
        tuple of (atom_center, family, nrot, aL, ops_car_sym, order_ops).
    """
    if type(poscar) == str:
        atom = read(poscar)
    elif type(poscar) == Atom or type(poscar) == Atoms:
        atom = poscar
    else:
        print("Unknown input")

    atom_center = find_axis_center_of_nanotube(atom)
    obj = LineGroupAnalyzer(atom_center, tolerance=1e-2)
    cyclic = CyclicGroupAnalyzer(atom_center, tolerance=1e-2)
    nrot = obj.get_rotational_symmetry_number()
    aL = atom_center.cell[2, 2]
    trans_sym = cyclic.cyclic_group[0]
    rota_sym = obj.sch_symbol

    trans_op = np.round(cyclic.get_generators(), 6)
    rots_op = np.round(obj.get_generators(), 6)

    if rots_op.size != 0:
        mats = np.vstack(([trans_op], rots_op))
    else:
        mats = trans_op.copy()
    ops, order_ops = brute_force_generate_group_subsequent(mats, symec=1e-2)

    ops_car_sym = []
    for op in ops:
        tmp_sym = SymmOp.from_rotation_and_translation(
            op[:3, :3], op[:3, 3] * aL
        )
        ops_car_sym.append(tmp_sym)
    family = get_family_num_from_sym_symbol(trans_sym, rota_sym)
    return atom_center, family, nrot, aL, ops_car_sym, order_ops


def main():
    """CLI entry point for computing irreps tables and character tables."""
    parser = argparse.ArgumentParser(
        description="Return the representation matrices or character table from a structure"
    )
    parser.add_argument(
        "-p", "--POSCAR", help="path to the file of a structure"
    )
    parser.add_argument(
        "-q",
        "--qpoint_z",
        type=float,
        default=0.0,
        help="The qpoint in the periodic direction (z), from 0 to 1",
    )
    parser.add_argument(
        "-t",
        "--tolerance",
        type=float,
        default=1e-8,
        help="Tolerance for atomic positions",
    )
    parser.add_argument(
        "-s",
        "--savename_chara",
        default="characters",
        help="The filename of character",
    )
    parser.add_argument(
        "-r",
        "--enable_rep_matrix",
        action="store_true",
        help="Enable output of irreducible representation matrices",
    )

    args = parser.parse_args()
    st_name = args.POSCAR
    qpoint_z = args.qpoint_z
    symprec = args.tolerance
    chara_filename = args.savename_chara
    enable_rep_matrix = args.enable_rep_matrix

    atom = read(st_name)

    (
        atom_center,
        family,
        nrot,
        aL,
        ops_car_sym,
        order_ops,
    ) = get_linegroup_symmetry_dataset(atom)
    qp_normalized = qpoint_z / aL * 2 * np.pi
    DictParams = {
        "qpoints": qp_normalized,
        "nrot": nrot,
        "order": order_ops,
        "family": family,
        "a": aL,
    }

    if enable_rep_matrix:
        (
            characters,
            irreps_values,
            irreps_symbols,
        ) = get_character_num_withparities(DictParams, symprec=symprec)
        representation_mat, _, _ = get_character_withparities(
            DictParams, symprec=symprec
        )
        representation_mat_dict = {}
        for i, rep in enumerate(representation_mat):
            representation_mat_dict[f"D_irrep_{i}"] = rep

        np.savez(
            chara_filename,
            characters=characters,
            ireps_values=irreps_values,
            ireps_symbols=irreps_symbols,
            **representation_mat_dict,
        )

    else:

        (
            characters,
            irreps_values,
            irreps_symbols,
        ) = get_character_num_withparities(DictParams, symprec=symprec)

        np.savez(
            chara_filename,
            characters=characters,
            ireps_values=irreps_values,
            ireps_symbols=irreps_symbols,
        )


if __name__ == "__main__":
    main()



# Copyright 2023-2026 The PULGON Project Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied. See the License for the specific language governing
# permissions and limitations under the License.

import argparse
import itertools
import logging
import warnings
from fractions import Fraction
from typing import List, Optional, Tuple, Union

import ase
import numpy as np
from ase import Atoms
from ase.io import read
from pymatgen.core.operations import SymmOp
from pymatgen.util.coord import find_in_coord_list

from pulgon_tools.utils import (
    angle_between_points,
    decimal_places,
    refine_cell,
)

warnings.filterwarnings(
    "ignore",
    category=RuntimeWarning,
    message="invalid value encountered in divide",
)


class CyclicGroupAnalyzer:
    """Analyze the generalized translational group
    (cyclic group).

    The general outline of the algorithm is as follows:

    1. Find primitive cell and get the pure translation
       distance in z direction.
    2. Get the potential monomers and the corresponding
       translational vector.
    3. Detect the possible cyclic group:
       - If rotational symmetry exists, it's a screw-axis
         group TQ(f) generated by (CQ|f).
       - If mirror symmetry exists, it's a glide plane
         group T'(f) generated by (sigmaV|f).

    """

    def __init__(
        self,
        atom: ase.atoms.Atoms,
        tolerance: float = 0.001,
    ) -> None:
        """
        Args:
            atom: Line group structure to determine
                the generalized translational group
            tolerance: distance tolerance for symmetry detection

        """
        logging.debug(
            "--------------------start detecting generalized translational group"
        )
        alongOZ = self._check_if_along_OZ(atom)
        if not alongOZ:
            logging.error(
                "Error while detect cyclic group: The axis direction is not OZ"
            )
        else:
            self._symprec = tolerance
            self._layer_symprec = 0.05

            round_symprec = decimal_places(tolerance)
            self._round_symprec = round_symprec
            self._zaxis = np.array([0, 0, 1])

            self._atom = self._find_axis_center_of_nanotube(atom)
            self._primitive = self._find_primitive()
            self.supercell_mutiple = int(
                len(self._atom.numbers) / len(self._primitive.numbers)
            )
            self._pure_trans = self._primitive.cell[2, 2]
            self._analyze()

    def _analyze(self) -> None:
        """Detect all possible monomers and determine their cyclic groups.

        Sets self.cyclic_group, self.monomers, and self._sym_operations.
        """
        monomer, potential_trans = self._potential_translation()

        logging.debug("There are %d monomer candidates" % len(monomer))
        (
            self.cyclic_group,
            self.monomers,
            self._sym_operations,
        ) = self._get_translations(monomer, potential_trans)

    def _find_axis_center_of_nanotube(
        self, atom: ase.atoms.Atoms
    ) -> ase.atoms.Atoms:
        """Shift the center of mass to scaled position (x,y) = (0.5, 0.5).

        Args:
            atom: initial structure

        Returns: centralized structure

        """
        n_st = atom.copy()
        center = self._get_center_of_mass_periodic(atom)

        pos = (
            np.remainder(atom.get_scaled_positions() - center + 0.5, [1, 1, 1])
            @ atom.cell
        )
        atoms = Atoms(
            cell=n_st.cell,
            numbers=n_st.numbers,
            positions=pos,
        )
        return atoms

    def _center_of_nanotube(self, atom: ase.atoms.Atoms) -> ase.atoms.Atoms:
        """Shift the center of mass to scaled position (x,y) = (0, 0).

        Args:
            atom: initial structure

        Returns: centralized structure

        """
        n_st = atom.copy()
        center = self._get_center_of_mass_periodic(atom)
        pos = (
            np.remainder(atom.get_scaled_positions() - center, [1, 1, 1])
            @ atom.cell
        )
        atoms = Atoms(
            cell=n_st.cell,
            numbers=n_st.numbers,
            positions=pos,
        )
        return atoms

    def _get_center_of_mass_periodic(
        self, atom: ase.atoms.Atoms
    ) -> np.ndarray:
        """Compute the center of mass in scaled
        coordinates with periodic wrapping.

        Uses circular mean for x and y; standard mean for z.

        Args:
            atom: ASE Atoms object.

        Returns:
            np.ndarray of scaled center-of-mass coordinates [cx, cy, cz].
        """
        L = np.array([1, 1, 1])
        x = atom.get_scaled_positions()
        theta = 2.0 * np.pi * x / L
        mass = atom.get_masses()
        mtheta = (
            np.arctan2(
                (-np.sin(theta) * np.expand_dims(mass, axis=1)).sum(axis=0)
                / len(theta),
                (-np.cos(theta) * np.expand_dims(mass, axis=1)).sum(axis=0)
                / len(theta),
            )
            + np.pi
        )
        center = L * mtheta / 2.0 / np.pi

        tmp = atom.get_center_of_mass(scaled=True)
        center[2] = tmp[2]
        return center

    def _get_translations(
        self, monomer_atoms: ase.atoms.Atoms, potential_tans: list
    ) -> tuple:
        """

        Args:
            monomer_atoms: possible monomers
            potential_tans: translational distances in z direction of monomers

        Returns: cyclic groups and monomers

        """
        cyclic_group, mono, sym_op = [], [], []
        invariant_op = SymmOp(
            [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 0]]
        )
        for ii, monomer in enumerate(monomer_atoms):
            logging.debug("---Start detecting Item %d monomer" % (ii + 1))
            tran = potential_tans[ii]
            ind = int(np.round(1 / tran))

            if ind - 1 / tran > self._symprec:
                logging.error("Selecting wrong translational vector")
                continue

            if len(monomer) == len(self._primitive):
                logging.debug(
                    "The monomer is self._primitive, so this cyclic group is T"
                )
                cyclic_group.append("T")
                mono.append(monomer)
                sym_op.append(invariant_op)
            else:
                # detect rotation
                logging.debug(
                    "The scaled translational distance is %s " % (1 / ind)
                )
                logging.debug("Start detecting rotation")

                rotation, Q, tmp_sym_operations = self._detect_rotation(
                    monomer, tran * self._pure_trans, ind
                )
                if rotation:
                    logging.debug(
                        "Append rotational cyclic group, Q is 360/degree=%s"
                        % Q
                    )
                    cyclic_group.append(
                        "(C%s|T%s(%s))"
                        % (Q, ind, np.round(tran * self._pure_trans, 3))
                    )
                    mono.append(monomer)
                    tmp_sym_operations.insert(0, invariant_op)
                    sym_op.append(tmp_sym_operations)
                else:
                    logging.debug("None of the candidate rotations works")

                if (
                    ind == 2 and abs(tran - 0.5) < self._layer_symprec
                ):  # only 2 layer in primitive cell
                    # detect mirror
                    logging.debug(
                        "The scaled translational distance is 1/2,"
                        " start detecting mirror symmetry"
                    )

                    mirror, tmp_sym_operations = self._detect_mirror(
                        monomer, self._pure_trans / 2
                    )
                    if mirror:
                        logging.debug(
                            "Mirror plane exist, append to cyclic group"
                        )
                        cyclic_group.append(
                            "T'(%s)" % np.round(self._pure_trans / 2, 3)
                        )
                        mono.append(monomer)

                        sym_op.append([invariant_op, tmp_sym_operations])

                    else:
                        logging.debug("Mirror does not exist")
        return cyclic_group, mono, sym_op

    def _detect_possible_helical_angle(
        self, ind: int, monomer: ase.atoms.Atoms
    ) -> np.ndarray:
        """Compute candidate helical rotation angles between monomer layers.

        Args:
            ind: number of monomer layers in the primitive cell.
            monomer: ASE Atoms of the base monomer.

        Returns:
            np.ndarray of unique candidate angles in degrees.
        """
        scaled = monomer.get_scaled_positions()
        z = scaled[:, 2]

        min_z = np.min(z)
        min_mask = np.abs(z - min_z) < self._symprec
        points_start = (scaled[min_mask] @ self._primitive.cell)[:, :2]

        # find out the points on upper layer
        target_z = (min_z + 1 / ind) % 1.0

        prim_z = self._primitive.get_scaled_positions()[:, 2]
        itp = np.where(np.abs(prim_z - target_z) < self._symprec)[0]

        if len(itp) == 0:
            raise RuntimeError(f"No atoms found at target z={target_z:.4f}")

        point_end = (
            self._primitive.get_scaled_positions() @ self._primitive.cell
        )[itp][:, :2]
        center = ([0.5, 0.5, 0] @ self._primitive.cell)[:2]

        # calculate the angle
        arr = np.array(
            [
                angle_between_points(ps, center, pe)
                for ps in points_start
                for pe in point_end
            ]
        )

        angles = np.unique(np.round(arr, self._round_symprec))
        angles = angles[np.abs(angles) > self._symprec]

        return angles

    def _detect_rotation(
        self,
        monomer: ase.atoms.Atoms,
        tran: np.float64,
        ind: int,
    ) -> Tuple[bool, Union[Fraction, int], Optional[list]]:
        """Detect helical/rotational symmetry between
        monomer layers in a nanotube.

        For each candidate rotation angle, verifies that
        rotating the base monomer by `angle * layer` and
        translating by `tran * layer` maps onto existing
        atoms in the primitive cell. Tests both positive
        and negative rotation directions independently.

        Args:
            monomer: Base monomer (layer 0) extracted
                from the primitive cell.
            tran: Translational displacement (Cartesian,
                along z) per monomer layer.
            ind: Number of monomer layers contained in
                the primitive cell.

        Returns:
            A tuple of (found, Q, sym_ops) where:
                - found (bool): True if a valid rotational
                  symmetry is detected.
                - Q (Fraction | int): Rotational index,
                  defined as 360 / angle. Returns 1 if no
                  symmetry is found.
                - sym_ops (list[SymmOp] | None): Symmetry
                  operations (one per layer, for the valid
                  rotation direction). None if not found.

        Notes:
            - When ind == 1, no inter-layer rotation can
              be defined; returns False.
            - Positive and negative rotation directions
              are tested separately to avoid mixing
              inconsistent symmetry operations.
            - Q is computed via Fraction to avoid
              floating-point precision issues.
        """
        if ind == 1:
            return False, 1, None

        coords = self._primitive.get_scaled_positions() @ self._primitive.cell
        center = [0.5, 0.5, 0] @ self._primitive.cell
        pot_angle = self._detect_possible_helical_angle(ind, monomer)
        logging.debug("Candidate rotational degrees are: %s" % str(pot_angle))

        for test_ind in pot_angle:
            for sign, label in [(1, "pos"), (-1, "neg")]:
                sym_ops = []
                all_pass = True

                for layer in range(1, ind):
                    angle = sign * test_ind * layer
                    op = SymmOp.from_origin_axis_angle(
                        origin=center, axis=self._zaxis, angle=angle
                    )
                    op = SymmOp.from_rotation_and_translation(
                        op.rotation_matrix,
                        op.translation_vector + [0, 0, tran * layer],
                    )

                    layer_pass = all(
                        len(
                            tmp := find_in_coord_list(
                                coords,
                                op.operate(site.position),
                                self._symprec,
                            )
                        )
                        == 1
                        and self._primitive.numbers[tmp[0]] == site.number
                        for site in monomer
                    )

                    if not layer_pass:
                        all_pass = False
                        break
                    sym_ops.append(op)
                if all_pass:
                    Q = (
                        Fraction(360, int(round(test_ind))).limit_denominator()
                        if float(test_ind).is_integer()
                        else Fraction(360 / test_ind).limit_denominator(1000)
                    )
                    return True, Q, sym_ops

        return False, 1, None

    def _detect_mirror(
        self,
        monomer: ase.atoms.Atoms,
        tran: np.float64,
    ) -> Tuple[bool, Optional[SymmOp]]:
        """

        Args:
            monomer: monomer candidates
            tran: the translational distance of monomer candidates

        Returns: judge if the mirror symmetry exist

        """

        coords = self._primitive.positions
        diff_st_ind = np.array(
            [
                find_in_coord_list(coords, coord, self._symprec)
                # for coord in sites_pos
                for coord in monomer.positions
            ]
        )

        if diff_st_ind.ndim > 1:
            diff_st_ind = diff_st_ind.T[0]
        # diff_st: diff_st + monomer = primitive cell
        diff_st = self._primitive[
            np.setdiff1d(range(len(coords)), diff_st_ind)
        ]
        for itp1, itp2 in itertools.combinations_with_replacement(
            range(len(monomer)), 2
        ):
            s1, s2 = monomer[itp1], diff_st[itp2]

            if (
                s1.number == s2.number
                and (s1.position[2] + tran - s2.position[2]) < self._symprec
            ):
                normal = s1.position - s2.position
                normal[2] = 0
                op = SymmOp.reflection(
                    normal, origin=([0.5, 0.5, 0.5] @ self._primitive.cell)
                )
                op = SymmOp.from_rotation_and_translation(
                    op.rotation_matrix,
                    op.translation_vector + np.array([0, 0, tran]),
                )
                itp = []
                for site in monomer:
                    coord = op.operate(site.position)
                    tmp = find_in_coord_list(
                        diff_st.positions, coord, self._symprec
                    )
                    itp.append(
                        len(tmp) == 1
                        and diff_st.numbers[tmp[0]] == site.number
                    )
                if np.array(itp).all():
                    return True, op
        return False, None

    def _get_monomer_ind(
        self, z_round: np.ndarray, z_uniq: np.ndarray
    ) -> Tuple[List[np.ndarray], List[np.ndarray]]:
        monomer_ind = [np.where(z_round == val)[0] for val in z_uniq]

        monomer_ind_sum = []
        accumulated = np.array([], dtype=np.int32)
        for ind in monomer_ind:
            accumulated = np.sort(np.append(accumulated, ind))
            monomer_ind_sum.append(accumulated.copy())
        return monomer_ind, monomer_ind_sum

    def _potential_translation(self) -> tuple:
        """Generate the potential monomer and the scaled
        translational distance in z axis.

        Returns: possible monomers and translational distances

        """
        z = self._primitive.get_scaled_positions()[:, 2]
        z_round = np.round(z, self._round_symprec)

        z_uniq, counts = np.unique(z_round, return_counts=True)
        monomer_ind, monomer_ind_sum = self._get_monomer_ind(z_round, z_uniq)

        z_uniq_high_accurate = np.array(
            [z[tmp_ind].mean().item() for tmp_ind in monomer_ind]
        )

        potential_trans = np.append(
            (z_uniq_high_accurate - z_uniq_high_accurate[0])[1:], 1
        )

        translation, monomer = [], []
        for ii in range(len(z_uniq)):
            monomer_num = counts[: ii + 1].sum()
            # check the atomic number and layer number
            # of potential monomer; check the translational
            # distance corresponds to the layer numbers
            if (
                len(self._primitive) % monomer_num
                == 0  # atom number is multiple of monomer
                and len(z_uniq) % (ii + 1) == 0  #
                and abs(len(z_uniq) / (ii + 1) - 1 / potential_trans[ii])
                < self._layer_symprec
            ):
                if len(self._primitive) == monomer_num:
                    # if the monomer is the whole structure
                    monomer.append(self._primitive)
                    translation.append(1)
                else:
                    monomer.append(self._primitive[monomer_ind_sum[ii]])
                    translation.append(potential_trans[ii])
        return monomer, translation

    def _find_primitive(self) -> ase.atoms.Atoms:
        """Find the primitive cell of a 1D line-group
        structure along z-axis."""
        scaled_pos = self._atom.get_scaled_positions()
        x_y = scaled_pos[:, :2]
        z = scaled_pos[:, 2]

        # Find candidate matching atoms in xy-plane
        indices = find_in_coord_list(x_y[1:], x_y[0], atol=self._symprec)
        indices = indices + 1 if indices.size > 0 else np.array([], dtype=int)

        if len(indices) == 0:
            logging.debug("It's already a primitive cell.")
            return self._atom

        potential_z = (z[indices] - z[0]) % 1
        trans_z = []

        # Check which translation preserves the cell
        for dz in potential_z:
            shifted_pos = scaled_pos + np.array([0, 0, dz])
            pos1, num1 = refine_cell(
                shifted_pos, self._atom.numbers, symprec=self._round_symprec
            )
            pos2, num2 = refine_cell(
                scaled_pos, self._atom.numbers, symprec=self._round_symprec
            )
            if np.allclose(pos1, pos2, atol=self._symprec) and np.array_equal(
                num1, num2
            ):
                trans_z.append(dz)

        if not trans_z:
            logging.debug("No valid translation found. Already primitive.")
            return self._atom

        # Build new primitive cell
        pure_z = min(trans_z)
        atom_indices = np.where(z < pure_z - self._symprec)[0]
        cell = np.zeros((3, 3))
        cell[0] = self._atom.cell[0]
        cell[1] = self._atom.cell[1]
        cell[2] = [0, 0, pure_z * self._atom.cell[2, 2]]

        atom = Atoms(
            numbers=self._atom.numbers[atom_indices],
            positions=self._atom.positions[atom_indices],
            cell=cell,
        )
        logging.debug(f"Primitive cell found with z-translation {pure_z}.")
        return atom

    def _check_if_along_OZ(self, atom: ase.atoms.Atoms) -> bool:
        """Check whether the cell's periodic axis is aligned along z.

        Args:
            atom: ASE Atoms object.

        Returns:
            bool: True if the third cell vector is along z only.
        """
        if (
            np.isclose(atom.cell[2, :2], [0, 0]).all()
            and np.isclose(atom.cell[:2, 2], [0, 0]).all()
        ):
            return True
        else:
            return False

    def get_cyclic_group(self) -> tuple:
        """Return the detected cyclic group symbols and corresponding monomers.

        Returns:
            tuple of (cyclic_group, monomers).
        """
        return self.cyclic_group, self.monomers

    def get_cyclic_group_and_op(self) -> tuple:
        """Return cyclic group symbols, monomers, and symmetry operations.

        Returns:
            tuple of (cyclic_group, monomers, sym_operations).
        """
        return self.cyclic_group, self.monomers, self._sym_operations

    def get_generators(self) -> np.ndarray:
        """Return the generator affine matrix of the cyclic group.

        Returns:
            np.ndarray: 4x4 affine matrix of the generator operation.
        """
        if len(self._sym_operations) == 1:
            return self._sym_operations[0].affine_matrix
        else:
            op = SymmOp.from_rotation_and_translation(
                rotation_matrix=self._sym_operations[0][1].rotation_matrix,
                translation_vec=[
                    0,
                    0,
                    (
                        self._sym_operations[0][1].translation_vector
                        @ np.linalg.inv(self._atom.cell)
                        % 1
                    )[2],
                ],
            )
            return op.affine_matrix


def main():
    """CLI entry point for detecting the generalized translational group."""
    parser = argparse.ArgumentParser(
        description="Try to detect the generalized translational group of a line group structure"
    )
    parser.add_argument(
        "-p",
        "--POSCAR",
        help="path to the file from which coordinates will be read",
    )
    parser.add_argument(
        "-t",
        "--tolerance",
        default=1e-3,
        type=float,
        help="Tolerance for atomic positions",
    )
    parser.add_argument(
        "-o",
        "--enable_log",
        action="store_true",
        help="Enable the output of detection process",
    )
    args = parser.parse_args()

    flag_log = args.enable_log
    if flag_log:
        logging.basicConfig(
            level=logging.DEBUG,
            format="%(asctime)s - %(levelname)s - %(message)s",
        )

    st_name = args.POSCAR
    st = read(st_name)

    cyclic = CyclicGroupAnalyzer(st, tolerance=args.tolerance)
    cy, monomers = cyclic.get_cyclic_group()
    print("The generalized translational group symbol:")
    for ii, cg in enumerate(cy):
        print("Monomer %s:" % monomers[ii].symbols, cg)


if __name__ == "__main__":
    main()



# Copyright 2023-2026 The PULGON Project Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied. See the License for the specific language governing
# permissions and limitations under the License.

import numpy as np
import pytest

from pulgon_tools.structure_chirality import (
    bond_constraints_equations,
    cyl2car,
    generate_symcell_and_linegroup_elements,
    get_nanotube_from_n1n2,
    helical_group_analysis,
)


@pytest.fixture
def hex_lattice():
    """Standard MoS2 hexagonal lattice vectors
    from bond_length=2.43, delta_Z=1.57."""
    bond_length = 2.43
    delta_Z = 1.57
    L1 = np.sqrt(bond_length**2 - delta_Z**2) * np.sqrt(
        2 - 2 * np.cos(2 * np.pi / 3)
    )
    a1 = L1 * np.array([1, 0])
    a2 = np.array([L1 * np.cos(np.pi / 3), L1 * np.sin(np.pi / 3)])
    return a1, a2, L1, bond_length, delta_Z


class TestCyl2Car:
    """Test cylindrical to Cartesian coordinate conversion."""

    def test_basic_conversion(self):
        car = cyl2car([np.pi / 4, 2.0, 3.0])
        expected = np.array([np.sqrt(2), np.sqrt(2), 3.0])
        assert np.allclose(car, expected, atol=1e-10)

    def test_zero_angle(self):
        car = cyl2car([0, 5.0, 1.0])
        assert np.allclose(car, [5.0, 0.0, 1.0], atol=1e-10)

    def test_pi_half_angle(self):
        car = cyl2car([np.pi / 2, 3.0, 0.0])
        assert np.allclose(car, [0.0, 3.0, 0.0], atol=1e-10)

    def test_zero_radius(self):
        car = cyl2car([1.23, 0.0, 4.0])
        assert np.allclose(car, [0.0, 0.0, 4.0], atol=1e-10)


class TestHelicalGroupAnalysis:
    """Test helical group parameter computation."""

    def test_armchair_10_10(self, hex_lattice):
        a1, a2, L1, _, _ = hex_lattice
        q, f, r, R, Ch, t, t1, t2, n_gcd = helical_group_analysis(
            a1, a2, 10, 10, L1
        )
        assert q == 20
        assert n_gcd == 10
        assert t1 == -1 and t2 == 1
        assert R == 3
        assert np.isclose(t, 3.2125, atol=1e-3)
        assert np.isclose(r, 8.8557, atol=1e-3)

    def test_zigzag_10_0(self, hex_lattice):
        a1, a2, L1, _, _ = hex_lattice
        q, f, r, R, Ch, t, t1, t2, n_gcd = helical_group_analysis(
            a1, a2, 10, 0, L1
        )
        assert q == 20
        assert n_gcd == 10
        assert t1 == -1 and t2 == 2
        assert R == 3
        assert np.isclose(t, 5.5642, atol=1e-3)

    def test_chiral_6_3(self, hex_lattice):
        a1, a2, L1, _, _ = hex_lattice
        q, f, r, R, Ch, t, t1, t2, n_gcd = helical_group_analysis(
            a1, a2, 6, 3, L1
        )
        assert q == 42
        assert n_gcd == 3
        assert np.isclose(t, 14.7214, atol=1e-3)

    def test_q_equals_n_gcd_times_q_tilde(self, hex_lattice):
        """Verify q = n_gcd * q_tilde relationship."""
        a1, a2, L1, _, _ = hex_lattice
        for n1, n2 in [(10, 10), (10, 0), (9, 9), (6, 3)]:
            q, f, r, R, Ch, t, t1, t2, n_gcd = helical_group_analysis(
                a1, a2, n1, n2, L1
            )
            q_tilde = q // n_gcd
            assert np.isclose(f, t / q_tilde, atol=1e-10)

    def test_chiral_vector_is_unit(self, hex_lattice):
        a1, a2, L1, _, _ = hex_lattice
        _, _, _, _, Ch, _, _, _, _ = helical_group_analysis(a1, a2, 10, 10, L1)
        assert np.isclose(np.linalg.norm(Ch), 1.0, atol=1e-10)


class TestBondConstraintsEquations:
    """Test the bond constraint equations for fsolve."""

    def test_equal_distances_give_zero_eq1(self):
        """If distances to pos2 and pos3 are equal,
        eq1 (their difference) is zero."""
        r = 5.0
        pos1 = [0.0, r, 0.0]
        phi = 0.3
        pos2 = [phi, r, 0.0]
        pos3 = [-phi, r, 0.0]
        pos4 = [0.0, r, 1.0]
        residuals = bond_constraints_equations(
            [0, 0, 0], pos1, pos2, pos3, pos4, 2.0
        )
        # eq1 = dist(1,2) - dist(1,3) should be zero by symmetry
        assert np.isclose(residuals[0], 0.0, atol=1e-10)

    def test_returns_three_equations(self):
        pos = [0.0, 3.0, 0.0]
        result = bond_constraints_equations([0, 0, 0], pos, pos, pos, pos, 2.0)
        assert len(result) == 3


class TestGenerateSymcellAndLinegroupElements:
    """Test symmetry cell generation for MoS2-type nanotubes."""

    def test_symcell_has_three_atoms(self, hex_lattice):
        a1, a2, L1, bond_length, delta_Z = hex_lattice
        _, _, r, _, Ch, _, t1, t2, _ = helical_group_analysis(
            a1, a2, 10, 10, L1
        )
        pos_cyl, symbols = generate_symcell_and_linegroup_elements(
            a1,
            a2,
            Ch,
            t1,
            t2,
            r,
            bond_length,
            delta_Z,
            symbol1=42,
            symbol2=16,
        )
        assert pos_cyl.shape == (3, 3)
        assert len(symbols) == 3
        assert symbols[0] == 42
        assert symbols[1] == 16
        assert symbols[2] == 16

    def test_symcell_radii(self, hex_lattice):
        """Metal atom at r, chalcogens at r +/- delta_Z."""
        a1, a2, L1, bond_length, delta_Z = hex_lattice
        _, _, r, _, Ch, _, t1, t2, _ = helical_group_analysis(
            a1, a2, 10, 10, L1
        )
        pos_cyl, _ = generate_symcell_and_linegroup_elements(
            a1, a2, Ch, t1, t2, r, bond_length, delta_Z
        )
        # pos_cyl[:, 1] are the radii
        r_metal = pos_cyl[0, 1]
        assert np.isclose(r_metal, r, atol=1e-6)


class TestGetNanotubeFromN1N2:
    """Test full nanotube generation from chiral indices."""

    def test_armchair_10_10_atom_count(self):
        atom = get_nanotube_from_n1n2(10, 10, "Mo", "S", 2.43, 1.57)
        assert len(atom) == 60

    def test_zigzag_10_0_atom_count(self):
        atom = get_nanotube_from_n1n2(10, 0, "Mo", "S", 2.43, 1.57)
        assert len(atom) == 60

    def test_armchair_9_9_atom_count(self):
        atom = get_nanotube_from_n1n2(9, 9, "Mo", "S", 2.43, 1.57)
        assert len(atom) == 54

    def test_chiral_6_3_atom_count(self):
        atom = get_nanotube_from_n1n2(6, 3, "Mo", "S", 2.43, 1.57)
        assert len(atom) == 135

    def test_stoichiometry(self):
        """MoS2-type: n_chalcogen = 2 * n_metal."""
        atom = get_nanotube_from_n1n2(10, 10, "Mo", "S", 2.43, 1.57)
        symbols = atom.get_chemical_symbols()
        n_Mo = symbols.count("Mo")
        n_S = symbols.count("S")
        assert n_S == 2 * n_Mo

    def test_cell_z_equals_t(self):
        """Cell z-dimension should equal the translational period t."""
        atom = get_nanotube_from_n1n2(10, 10, "Mo", "S", 2.43, 1.57)
        assert np.isclose(atom.cell[2, 2], 3.2125, atol=1e-3)

    def test_bond_length_preserved(self):
        """Nearest Mo-S distance should match the input bond_length."""
        bond_length = 2.43
        atom = get_nanotube_from_n1n2(10, 10, "Mo", "S", bond_length, 1.57)
        pos = atom.positions
        symbols = atom.get_chemical_symbols()
        mo_idx = [i for i, s in enumerate(symbols) if s == "Mo"]
        s_idx = [i for i, s in enumerate(symbols) if s == "S"]
        min_d = min(
            np.linalg.norm(pos[i] - pos[j]) for i in mo_idx for j in s_idx
        )
        assert np.isclose(min_d, bond_length, atol=1e-3)

    def test_no_duplicate_atoms(self):
        """All atoms should be at unique positions."""
        atom = get_nanotube_from_n1n2(10, 10, "Mo", "S", 2.43, 1.57)
        scaled = atom.get_scaled_positions()
        unique = np.unique(np.round(scaled, 8), axis=0)
        assert len(unique) == len(atom)

    def test_different_symbols(self):
        """Support arbitrary element symbols."""
        atom = get_nanotube_from_n1n2(9, 9, "W", "Se", 2.43, 1.57)
        symbols = atom.get_chemical_symbols()
        assert "W" in symbols
        assert "Se" in symbols
        assert symbols.count("Se") == 2 * symbols.count("W")



# Copyright 2023-2026 The PULGON Project Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied. See the License for the specific language governing
# permissions and limitations under the License.

import numpy as np
from ase.io.vasp import read_vasp
from pymatgen.core import Molecule
from pymatgen.core.operations import SymmOp

from pulgon_tools.detect_generalized_translational_group import (
    CyclicGroupAnalyzer,
)
from pulgon_tools.detect_point_group import LineGroupAnalyzer, get_symcell

pytest_plugins = ["pytest-datadir"]


class TestCyclicGroupAnalyzer:
    def test_rotation(self, shared_datadir):
        st_name = shared_datadir / "st1"
        st = read_vasp(st_name)
        cy = CyclicGroupAnalyzer(st, tolerance=1e-2)
        monomers, translations = cy._potential_translation()
        idx, Q, _ = cy._detect_rotation(
            monomers[0], translations[0] * cy._primitive.cell[2, 2], 3
        )

        assert idx == True
        assert Q.numerator == 12

    def test_find_axis_center_of_nanotube(self, shared_datadir):
        st_name = shared_datadir / "12-12-AM"
        st = read_vasp(st_name)
        res = CyclicGroupAnalyzer(st)
        n_st = res._find_axis_center_of_nanotube(st)
        average_coord = (n_st.positions[:, :2] / len(n_st)).sum(axis=0)
        assert (average_coord - [0.5, 0.5] @ n_st.cell[:2, :2]).sum() < 0.001

    def test_generate_monomer(self, shared_datadir):
        st_name = shared_datadir / "9-9-AM"
        st = read_vasp(st_name)
        cy = CyclicGroupAnalyzer(st)
        monomers, translations = cy._potential_translation()
        assert str(monomers[0].symbols) == "Mo9S18"
        assert np.isclose(translations[0], 0.5, atol=1e-3)

    def test_rotational_tolerance(self, shared_datadir):
        st_name = shared_datadir / "st1"
        st = read_vasp(st_name)
        cy1 = CyclicGroupAnalyzer(st, tolerance=1e-2)
        cy2 = CyclicGroupAnalyzer(st, tolerance=1e-4)
        monomers1, translations1 = cy1._potential_translation()
        monomers2, translations2 = cy2._potential_translation()
        idx1, _, _ = cy1._detect_rotation(
            monomers1[0], translations1[0] * cy1._primitive.cell[2, 2], ind=3
        )
        idx2, _, _ = cy2._detect_rotation(
            monomers2[0], translations2[0] * cy2._primitive.cell[2, 2], ind=3
        )
        assert idx1 == True
        assert idx2 == False

    def test_mirror(self, shared_datadir):
        st_name = shared_datadir / "st7"
        st = read_vasp(st_name)
        cy = CyclicGroupAnalyzer(st)
        monomers, translations = cy._potential_translation()
        idx, _ = cy._detect_mirror(
            monomers[0], translations[0] * cy._primitive.cell[2, 2]
        )
        assert idx == True

    def test_mirror_tolerance(self, shared_datadir):
        st_name = shared_datadir / "st7"
        st = read_vasp(st_name)

        cy1 = CyclicGroupAnalyzer(st, tolerance=1e-13)
        cy2 = CyclicGroupAnalyzer(st, tolerance=1e-16)
        monomers1, translations1 = cy1._potential_translation()
        monomers2, translations2 = cy2._potential_translation()
        idx1, _ = cy1._detect_mirror(
            monomers1[0], translations1[0] * cy1._primitive.cell[2, 2]
        )
        idx2, _ = cy2._detect_mirror(
            monomers2[0], translations2[0] * cy2._primitive.cell[2, 2]
        )

        assert idx1 == True
        assert idx2 == False

    def test_the_whole_function_st1(self, shared_datadir):
        st_name = shared_datadir / "st1"
        st = read_vasp(st_name)
        cyclic = CyclicGroupAnalyzer(st, tolerance=1e-2)
        cy, mon = cyclic.get_cyclic_group()
        assert cy[0] == "(C12|T3(1.5))" and str(mon[0].symbols) == "C4"

    def test_the_whole_function_st2(self, shared_datadir):
        st_name = shared_datadir / "st7"
        st = read_vasp(st_name)
        cyclic = CyclicGroupAnalyzer(st, tolerance=1e-2)
        cy, mon = cyclic.get_cyclic_group()
        assert cy[0] == "(C24|T2(1.5))" and str(mon[0].symbols) == "C6"

    def test_the_whole_function_st3(self, shared_datadir):
        st_name = shared_datadir / "9-9-AM"
        st = read_vasp(st_name)
        cyclic = CyclicGroupAnalyzer(st, tolerance=1e-2)
        cy, mon = cyclic.get_cyclic_group()
        assert cy[0] == "(C18|T2(1.606))" and str(mon[0].symbols) == "Mo9S18"

    def test_the_whole_function_st4(self, shared_datadir):
        st_name = shared_datadir / "24-0-ZZ"
        st = read_vasp(st_name)
        cyclic = CyclicGroupAnalyzer(st, tolerance=1e-2)
        cy, mon = cyclic.get_cyclic_group()
        assert (
            cy[0] == "(C48|T2(2.74))"
            and cy[1] == "T'(2.74)"
            and str(mon[0].symbols) == "Mo24S48"
        )


class TestAxialPointGroupAnalyzer:
    def test_axial_pg_st1(self, shared_datadir):
        st_name = shared_datadir / "m1"
        st = read_vasp(st_name)
        mol = Molecule(species=st.numbers, coords=st.positions)
        obj = LineGroupAnalyzer(mol)
        pg = obj.get_pointgroup()
        assert str(pg) == "D6h"

    def test_axial_pg_st2(self, shared_datadir):
        st_name = shared_datadir / "m2"
        st = read_vasp(st_name)
        mol = Molecule(species=st.numbers, coords=st.positions)
        obj = LineGroupAnalyzer(mol)
        pg = obj.get_pointgroup()
        assert str(pg) == "D6d"

    def test_axial_pg_st3(self, shared_datadir):
        st_name = shared_datadir / "m4"
        st = read_vasp(st_name)
        mol = Molecule(species=st.numbers, coords=st.positions)
        obj = LineGroupAnalyzer(mol)
        pg = obj.get_pointgroup()
        assert str(pg) == "D6"

    def test_axial_pg_st4(self, shared_datadir):
        st_name = shared_datadir / "9-9-AM"
        st = read_vasp(st_name)
        mol = Molecule(species=st.numbers, coords=st.positions)
        obj = LineGroupAnalyzer(mol)
        pg = obj.get_pointgroup()
        assert str(pg) == "S18"

    def test_axial_pg_st6(self, shared_datadir):
        st_name = shared_datadir / "C4h"
        st = read_vasp(st_name)
        mol = Molecule(species=st.numbers, coords=st.positions)
        obj = LineGroupAnalyzer(mol)
        pg = obj.get_pointgroup()
        assert str(pg) == "C4h"

    def test_axial_pg_st7(self, shared_datadir):
        st_name = shared_datadir / "C4v"
        st = read_vasp(st_name)
        mol = Molecule(species=st.numbers, coords=st.positions)
        obj = LineGroupAnalyzer(mol)
        pg = obj.get_pointgroup()
        assert str(pg) == "C4v"

    def test_axial_pg_st8(self, shared_datadir):
        st_name = shared_datadir / "C4"
        st = read_vasp(st_name)
        mol = Molecule(species=st.numbers, coords=st.positions)
        obj = LineGroupAnalyzer(mol)
        pg = obj.get_pointgroup()
        assert str(pg) == "C4"

    def test_axial_pg_st9(self, shared_datadir):
        st_name = shared_datadir / "non-sym"
        st = read_vasp(st_name)
        mol = Molecule(species=st.numbers, coords=st.positions)
        obj = LineGroupAnalyzer(mol)
        pg = obj.get_pointgroup()
        assert str(pg) == "C1"

    def test_axial_pg_st10(self, shared_datadir):
        st_name = shared_datadir / "24-0-ZZ"
        st = read_vasp(st_name)
        mol = Molecule(species=st.numbers, coords=st.positions)
        obj = LineGroupAnalyzer(st)
        pg = obj.get_pointgroup()
        assert str(pg) == "C24v"


class TestCyclicGroupAnalyzerMethods:
    """Test individual methods of CyclicGroupAnalyzer."""

    def test_get_cyclic_group_and_op(self, shared_datadir):
        """get_cyclic_group_and_op returns groups, monomers, and operations."""
        st = read_vasp(shared_datadir / "st7")
        cy = CyclicGroupAnalyzer(st, tolerance=1e-2)
        groups, monomers, ops = cy.get_cyclic_group_and_op()
        assert len(groups) == len(monomers) == len(ops)
        assert len(groups) > 0

    def test_get_generators_screw_axis(self, shared_datadir):
        """Generator for a screw-axis group should be a 4x4 affine matrix."""
        st = read_vasp(shared_datadir / "9-9-AM")
        cy = CyclicGroupAnalyzer(st, tolerance=1e-2)
        gen = cy.get_generators()
        assert gen.shape == (4, 4)
        # Should contain a rotation (not identity)
        assert not np.allclose(gen[:3, :3], np.eye(3))

    def test_get_generators_pure_translation(self, shared_datadir):
        """Generator for a T-only group should be 4x4 affine matrix."""
        st = read_vasp(shared_datadir / "C4")
        cy = CyclicGroupAnalyzer(st, tolerance=1e-2)
        gen = cy.get_generators()
        assert gen.shape == (4, 4)

    def test_center_of_nanotube(self, shared_datadir):
        """_center_of_nanotube returns a valid centered structure."""
        st = read_vasp(shared_datadir / "12-12-AM")
        cy = CyclicGroupAnalyzer(st, tolerance=1e-2)
        centered = cy._center_of_nanotube(st)
        assert len(centered) == len(st)
        assert centered.cell is not None

    def test_check_if_along_OZ_true(self, shared_datadir):
        st = read_vasp(shared_datadir / "st1")
        cy = CyclicGroupAnalyzer(st, tolerance=1e-2)
        assert cy._check_if_along_OZ(st) is True

    def test_check_if_along_OZ_false(self):
        """A cell with off-diagonal z-components should fail the OZ check."""
        from ase import Atoms

        bad_atom = Atoms(
            "C2",
            positions=[[0, 0, 0], [1, 1, 1]],
            cell=[[10, 0, 0], [0, 10, 0], [1, 0, 5]],
        )
        # Need an existing CyclicGroupAnalyzer instance to call the method
        # Use a valid structure to create one, then test with bad cell
        bad_atom2 = Atoms(
            "C2",
            positions=[[0, 0, 0], [1, 1, 1]],
            cell=[[10, 0, 0], [0, 10, 0], [0, 0, 5]],
        )
        cy = CyclicGroupAnalyzer(bad_atom2, tolerance=1e-2)
        assert cy._check_if_along_OZ(bad_atom) is False

    def test_detect_rotation_ind1_returns_false(self, shared_datadir):
        """When ind==1, _detect_rotation should return False immediately."""
        st = read_vasp(shared_datadir / "st1")
        cy = CyclicGroupAnalyzer(st, tolerance=1e-2)
        monomers, translations = cy._potential_translation()
        found, Q, ops = cy._detect_rotation(
            monomers[0], translations[0] * cy._primitive.cell[2, 2], ind=1
        )
        assert found is False
        assert Q == 1
        assert ops is None

    def test_find_primitive_already_primitive(self, shared_datadir):
        """_find_primitive returns unchanged when
        the structure is already primitive."""
        st = read_vasp(shared_datadir / "C4")
        cy = CyclicGroupAnalyzer(st, tolerance=1e-2)
        assert len(cy._primitive) == len(cy._atom)
        assert cy.supercell_mutiple == 1


class TestLineGroupAnalyzerMethods:
    """Test individual methods of LineGroupAnalyzer."""

    def test_get_symmetry_operations_D6h(self, shared_datadir):
        """D6h should have 24 symmetry operations."""
        st = read_vasp(shared_datadir / "m1")
        mol = Molecule(species=st.numbers, coords=st.positions)
        obj = LineGroupAnalyzer(mol)
        ops = obj.get_symmetry_operations()
        assert len(ops) == 24
        assert all(isinstance(op, SymmOp) for op in ops)

    def test_get_symmetry_operations_C1(self, shared_datadir):
        """C1 should have only the identity."""
        st = read_vasp(shared_datadir / "non-sym")
        mol = Molecule(species=st.numbers, coords=st.positions)
        obj = LineGroupAnalyzer(mol)
        ops = obj.get_symmetry_operations()
        assert len(ops) == 1

    def test_get_generators_D6h(self, shared_datadir):
        st = read_vasp(shared_datadir / "m1")
        mol = Molecule(species=st.numbers, coords=st.positions)
        obj = LineGroupAnalyzer(mol)
        gens = obj.get_generators()
        assert len(gens) > 0
        for g in gens:
            assert g.shape == (4, 4)

    def test_get_generators_excludes_identity(self, shared_datadir):
        st = read_vasp(shared_datadir / "m1")
        mol = Molecule(species=st.numbers, coords=st.positions)
        obj = LineGroupAnalyzer(mol)
        gens = obj.get_generators()
        for g in gens:
            assert not np.allclose(g, np.eye(4))

    def test_inertia_tensor_symmetric(self, shared_datadir):
        st = read_vasp(shared_datadir / "m1")
        mol = Molecule(species=st.numbers, coords=st.positions)
        obj = LineGroupAnalyzer(mol)
        I = obj._inertia_tensor()
        assert I.shape == (3, 3)
        assert np.allclose(I, I.T, atol=1e-10)

    def test_center_of_mass_periodic(self, shared_datadir):
        st = read_vasp(shared_datadir / "12-12-AM")
        mol = Molecule(species=st.numbers, coords=st.positions)
        obj = LineGroupAnalyzer(mol)
        com = obj._get_center_of_mass_periodic(st)
        assert com.shape == (3,)
        assert np.all(com >= 0) and np.all(com <= 1)

    def test_find_axis_center(self, shared_datadir):
        st = read_vasp(shared_datadir / "12-12-AM")
        mol = Molecule(species=st.numbers, coords=st.positions)
        obj = LineGroupAnalyzer(mol)
        centered = obj._find_axis_center_of_nanotube(st)
        assert len(centered) == len(st)
        # After centering, the mean xy position should be near cell center
        scaled = centered.get_scaled_positions()
        mean_xy = scaled[:, :2].mean(axis=0)
        assert np.allclose(mean_xy, [0.5, 0.5], atol=0.05)

    def test_accepts_atoms_object(self, shared_datadir):
        """LineGroupAnalyzer should accept ASE Atoms directly."""
        st = read_vasp(shared_datadir / "12-12-AM")
        obj = LineGroupAnalyzer(st, tolerance=0.01)
        pg = obj.get_pointgroup()
        assert str(pg) != ""


class TestGetSymcell:
    """Test get_symcell function."""

    def test_symcell_reduces_atoms(self, shared_datadir):
        """Symmetry cell should have fewer atoms than the full monomer."""
        st = read_vasp(shared_datadir / "m1")
        symcell = get_symcell(st)
        assert len(symcell) < len(st)

    def test_symcell_nonempty(self, shared_datadir):
        st = read_vasp(shared_datadir / "m1")
        symcell = get_symcell(st)
        assert len(symcell) >= 1



# Copyright 2023-2026 The PULGON Project Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied. See the License for the specific language governing
# permissions and limitations under the License.


import argparse
import ast
import collections
import os
from typing import List, Optional, Tuple, Union

import cvxpy as cp
import matplotlib.pyplot as plt
import numpy as np
import phonopy.file_IO
import scipy as sp
import scipy.sparse as ssp
from ase import Atoms
from ase.io import read
from phonopy import Phonopy
from phonopy.file_IO import parse_FORCE_CONSTANTS, read_force_constants_hdf5
from phonopy.phonon.band_structure import get_band_qpoints_and_path_connections
from phonopy.structure.atoms import PhonopyAtoms
from sklearn.linear_model import Ridge


def calc_dists(
    atoms: Atoms, tolerance: float = 1e-4
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Return the distances between atoms in the supercell, their
    degeneracies and the associated displacements along OZ.
    """
    MIN_DELTA = -2
    MAX_DELTA = 2
    positions = atoms.positions
    cell = atoms.cell
    n_satoms = positions.shape[0]
    d2s = np.empty((MAX_DELTA - MIN_DELTA + 1, n_satoms, n_satoms))
    # TODO: This could not be enough and eventually we should do a proper check
    # for the shortest distance.
    for j, j_c in enumerate(range(MIN_DELTA, MAX_DELTA + 1)):
        shifted_positions = positions + (j_c * cell[2, :])[np.newaxis, :]
        d2s[j, :, :] = sp.spatial.distance.cdist(
            positions, shifted_positions, "sqeuclidean"
        )
    d2min = d2s.min(axis=0)
    dmin = np.sqrt(d2min)
    degenerate = np.abs(d2s - d2min) < tolerance
    nequi = degenerate.sum(axis=0, dtype=int)
    maxequi = nequi.max()
    sorting = np.argsort(np.logical_not(degenerate), axis=0)
    shifts = np.transpose(sorting[:maxequi, :, :], (1, 2, 0)).astype(np.intc)
    shifts = np.asarray(range(MIN_DELTA, MAX_DELTA + 1))[shifts]
    return (dmin, nequi, shifts)


def build_constraint_matrix(
    phonon: Phonopy, recenter: bool = False
) -> Tuple[ssp.spmatrix, np.ndarray]:
    """
    Build the sparse constraint matrix encoding translational
    acoustic sum rules, Born-Huang rotational sum rules,
    Huang invariances, matrix symmetry, and cutoff constraints.
    Returns: sparse matrix M, IFC array
    """
    scell = phonon.supercell
    if recenter:
        scell.scaled_positions = (scell.scaled_positions - [0.5, 0.5, 0.5]) % 1

    phonon.symmetrize_force_constants()
    IFC = phonon.force_constants.copy()
    if IFC.shape[0] == IFC.shape[1]:
        p2s_map = phonon.primitive.p2s_map
        IFC = IFC[p2s_map]  # return to compact format

    motif_indices = [
        phonon.primitive.p2p_map[i] for i in phonon.primitive.s2p_map
    ]
    supercell_indices = []
    counter = collections.Counter()
    for i_atom in motif_indices:
        supercell_indices.append(counter[i_atom])
        counter[i_atom] += 1
    n_cells = max(supercell_indices) + 1

    symbols = scell.symbols
    cell = scell.cell

    positions = (
        (scell.scaled_positions + np.asarray([0.0, 0.0, 0.0])[np.newaxis, :])
        % 1.0
    ) @ cell
    ase_atoms = Atoms(symbols, positions, cell=cell)

    dists, degeneracy, shifts = calc_dists(ase_atoms)

    n_atoms, n_satoms = IFC.shape[:2]

    average_delta = np.zeros((n_satoms, n_satoms, 3))
    for i in range(n_satoms):
        for j in range(n_satoms):
            n_elements = degeneracy[i, j]
            for i_d in range(n_elements):
                average_delta[i, j, :] += (
                    positions[j, :]
                    - positions[i, :]
                    + shifts[i, j, i_d] * cell[2, :]
                )
            average_delta[i, j, :] /= n_elements

    average_products = np.zeros((n_satoms, n_satoms, 3, 3))
    for i in range(n_satoms):
        for j in range(n_satoms):
            n_elements = degeneracy[i, j]
            for i_d in range(n_elements):
                delta = (
                    positions[j, :]
                    - positions[i, :]
                    + shifts[i, j, i_d] * cell[2, :]
                )
                average_products[i, j, :, :] += np.outer(delta, delta)
            average_products[i, j, :, :] /= n_elements

    n_rows = 0
    rows = []
    cols = []
    data = []
    # Acoustic sum rules for translations.
    for i in range(n_atoms):
        for alpha in range(3):
            for beta in range(3):
                for j in range(n_satoms):
                    rows.append(n_rows)
                    cols.append(
                        np.ravel_multi_index((i, j, alpha, beta), IFC.shape)
                    )
                    data.append(1.0)
                n_rows += 1
    print("Adding acoustic sum rules")

    # The same but for rotations (Born-Huang).
    for i in range(n_atoms):
        for alpha in range(3):
            for beta in range(3):
                for gamma in range(3):
                    for j in range(n_satoms):
                        r_ij = average_delta[phonon.primitive.p2s_map[i], j]
                        rows.append(n_rows)
                        cols.append(
                            np.ravel_multi_index(
                                (i, j, alpha, beta), IFC.shape
                            )
                        )
                        data.append(r_ij[gamma])
                        rows.append(n_rows)
                        cols.append(
                            np.ravel_multi_index(
                                (i, j, alpha, gamma), IFC.shape
                            )
                        )
                        data.append(-r_ij[beta])
                    n_rows += 1
    print("Adding Born-Huang sum rules")

    # And the Huang invariances, also for rotation.
    for alpha in range(3):
        for beta in range(3):
            for gamma in range(3):
                for delta in range(3):
                    for i in range(n_atoms):
                        for j in range(n_satoms):
                            products = average_products[
                                phonon.primitive.p2s_map[i], j
                            ]
                            rows.append(n_rows)
                            cols.append(
                                np.ravel_multi_index(
                                    (i, j, alpha, beta), IFC.shape
                                )
                            )
                            data.append(products[gamma, delta])
                            rows.append(n_rows)
                            cols.append(
                                np.ravel_multi_index(
                                    (i, j, gamma, delta), IFC.shape
                                )
                            )
                            data.append(-products[alpha, beta])
                    n_rows += 1
    print("Adding Huang invariances")

    # Make sure the IFC matrix is symmetric.
    for i in range(n_atoms):
        for j in range(n_atoms):
            for alpha in range(3):
                for beta in range(3):
                    rows.append(n_rows)
                    cols.append(
                        np.ravel_multi_index(
                            (i, phonon.primitive.p2s_map[j], alpha, beta),
                            IFC.shape,
                        )
                    )
                    data.append(1.0)
                    rows.append(n_rows)
                    cols.append(
                        np.ravel_multi_index(
                            (j, phonon.primitive.p2s_map[i], beta, alpha),
                            IFC.shape,
                        )
                    )
                    data.append(-1.0)
                    n_rows += 1
    print("now finish symmetric rules")

    # Make the tensor symmetric in a PBC setting (H01 = transpose(H10)).
    for i in range(n_atoms):
        for j in range(n_atoms):
            for alpha in range(3):
                for beta in range(3):
                    rows.append(n_rows)

                    j_in_second_cell = [
                        m for m, n in enumerate(motif_indices) if n == j
                    ][1]
                    cols.append(
                        np.ravel_multi_index(
                            (i, j_in_second_cell, alpha, beta), IFC.shape
                        )
                    )
                    data.append(1.0)
                    rows.append(n_rows)
                    i_in_last_cell = [
                        m for m, n in enumerate(motif_indices) if n == i
                    ][-1]
                    cols.append(
                        np.ravel_multi_index(
                            (j, i_in_last_cell, beta, alpha), IFC.shape
                        )
                    )
                    data.append(-1.0)
                    n_rows += 1
    print("now finish the (H01 = transpose(H10) constrains")

    # Add extra constraints to make the force constants
    # short-sighted for the cell
    # transmission calculations.
    for j in range(n_satoms):
        if supercell_indices[j] not in (0, 1, n_cells - 1):
            for i in range(n_atoms):
                for alpha in range(3):
                    for beta in range(3):
                        rows.append(n_rows)
                        cols.append(
                            np.ravel_multi_index(
                                (i, j, alpha, beta), IFC.shape
                            )
                        )
                        data.append(1.0)
                        n_rows += 1
    print("now finish the short-sighted(the cell) constrains")

    # Rebuild the sparse matrix.
    M = ssp.coo_array((data, (rows, cols)), shape=(n_rows, IFC.size))
    return M, IFC


def solve_fcs(
    IFC: np.ndarray,
    M: ssp.spmatrix,
    methods: str = "convex_opt",
) -> np.ndarray:
    """
    Solve the constrained quadratic optimization problem
    to correct the IFCs.
    Returns: corrected IFC array
    """

    if methods == "convex_opt":

        # CVXPY convex optimization
        flat_IFCs = IFC.ravel()
        print("M @ x  before:", np.abs(M.dot(flat_IFCs)).sum())
        x = cp.Variable(IFC.size)
        cost = cp.sum_squares(x - flat_IFCs)
        prob = cp.Problem(cp.Minimize(cost), [M @ x == 0])
        prob.solve()
        IFC_sym = x.value.reshape(IFC.shape)
        print("M @ x  after:", np.abs(M.dot(IFC_sym.ravel())).sum())

    elif methods == "ridge_model":

        # Ridge regression
        parameters = IFC.ravel().copy()
        d = M.dot(parameters)
        delta = np.linalg.norm(d)
        print("Rotational sum-rules before, ||Ax|| = {:20.15f}".format(delta))
        ridge = Ridge(
            alpha=1e-6, fit_intercept=False, solver="sparse_cg", tol=1e-10
        )
        ridge.fit(M, d)
        parameters -= ridge.coef_
        d = M.dot(parameters)
        delta = np.linalg.norm(d)
        print("Rotational sum-rules after,  ||Ax|| = {:20.15f}".format(delta))
        IFC_sym = parameters.reshape(IFC.shape)
    return IFC_sym


def main():
    """CLI entry point for correcting force constants with sum rules."""
    parser = argparse.ArgumentParser(description="Apply the sum rules to fcs")
    parser.add_argument(
        "-p",
        "--POSCAR",
        default="POSCAR",
        help="The path of poscar",
    )
    parser.add_argument(
        "-x",
        "--supercell_matrix",
        required=True,
        # default=[5,5,1],
        type=parse_int_list,
        help="The supercell matrix that used to calculate fcs",
    )
    parser.add_argument(
        "-y",
        "--path_yaml",
        default=None,
        help="The path of phonopy.yaml",
    )
    parser.add_argument(
        "-f",
        "--fcs",
        default="./FORCE_CONSTANTS",
        help="The path of force_constants.hdf5 or FORCE_CONSTANTS",
    )
    parser.add_argument(
        "-n",
        "--plot_phonon",
        action="store_true",
        help="Enable plotting if specified (default: False)",
    )
    parser.add_argument(
        "-k",
        "--k_path",
        default=None,
        type=str2list,
        help="The k path of plotting phonon, e.g. [[0.0, 0.0, 0.0], [0.5, 0.0, 0.0], [0.0, 0.0, 0.0]]",
    )
    parser.add_argument(
        "-r",
        "--recenter",
        action="store_true",
        help="(atoms.positions - [0.5,0.5,0.5])%1",
    )
    parser.add_argument(
        "-m",
        "--methods",
        default="convex_opt",
        help="The available methods are 'convex_opt', 'ridge_model'",
    )

    args = parser.parse_args()

    supercell_matrix = args.supercell_matrix
    path_yaml = args.path_yaml
    fcs_name = args.fcs
    methods = args.methods
    plot_phonon = args.plot_phonon
    recenter = args.recenter
    k_path = args.k_path
    fcs_savename = "FORCE_CONSTANTS_correction"
    phononfig_savename = "phonon_fix"

    if path_yaml is not None:
        phonon = phonopy.load(
            phonopy_yaml=path_yaml, force_constants_filename=fcs_name
        )
    else:
        poscar = args.POSCAR
        atoms = read(poscar)

        unitcell = PhonopyAtoms(
            symbols=atoms.get_chemical_symbols(),
            cell=atoms.cell.array,
            scaled_positions=atoms.get_scaled_positions(),
        )

        phonon = Phonopy(unitcell, supercell_matrix=np.diag(supercell_matrix))
        file_ext = os.path.splitext(fcs_name)[1].lower()
        if file_ext == ".hdf5":
            fcs = read_force_constants_hdf5(fcs_name)
        else:
            fcs = parse_FORCE_CONSTANTS(fcs_name)

        phonon.force_constants = fcs

    IFC_raw_backup = phonon.force_constants.copy()

    M, IFC = build_constraint_matrix(phonon, recenter=recenter)

    print("Start solving constraints")
    IFC_sym = solve_fcs(IFC, M, methods=methods)

    if plot_phonon:
        print("Start drawing the phonon spectrum")
        if k_path is None:
            phonon.force_constants = IFC_raw_backup

            phonon.auto_band_structure()
            phonon.plot_band_structure().savefig(phononfig_savename, dpi=500)
        else:
            qpoints, connections = get_band_qpoints_and_path_connections(
                [k_path], npoints=101
            )

            phonon.force_constants = IFC_raw_backup
            phonon.run_band_structure(
                qpoints, path_connections=connections, with_eigenvectors=True
            )
            bands_raws = phonon.get_band_structure_dict()

            phonon.force_constants = IFC_sym
            phonon.run_band_structure(
                qpoints, path_connections=connections, with_eigenvectors=True
            )
            bands_fix = phonon.get_band_structure_dict()

            distances = np.concatenate(bands_raws["distances"])
            freq_raw = np.concatenate(bands_raws["frequencies"])
            freq_fix = np.concatenate(bands_fix["frequencies"])

            for ii, freq in enumerate(freq_raw.T):
                if ii == 0:
                    plt.plot(distances, freq, label="raw", color="grey")
                else:
                    plt.plot(distances, freq, color="grey")

            for ii, freq in enumerate(freq_fix.T):
                if ii == 0:
                    plt.plot(distances, freq, label="fix", color="red")
                else:
                    plt.plot(distances, freq, color="red")
            plt.plot(distances, np.zeros_like(distances), "b--", linewidth=0.8)
            plt.legend()
            # plt.ylim(bottom=0)
            plt.savefig(phononfig_savename, dpi=300)

    phonopy.file_IO.write_FORCE_CONSTANTS(
        IFC_sym, fcs_savename, phonon.primitive.p2s_map
    )
    phonopy.file_IO.write_force_constants_to_hdf5(
        IFC_sym, filename=(fcs_savename + ".hdf5")
    )


def parse_bool_list(value: str) -> List[bool]:
    """Parse a string representing a list of booleans into a Python list"""
    try:
        parsed = ast.literal_eval(value)
        if not isinstance(parsed, list) or not all(
            isinstance(x, bool) for x in parsed
        ):
            raise ValueError(
                "Input must be a list of booleans, e.g., [True, True, False]"
            )
        if len(parsed) != 3:
            raise ValueError(
                "PBC list must contain exactly 3 booleans, e.g., [True, True, False]"
            )
        return parsed
    except (ValueError, SyntaxError):
        raise argparse.ArgumentTypeError(f"Invalid bool list: {value}")


def parse_int_list(value: str) -> List[int]:
    """Parse a string representing a list of 3 integers into a Python list."""
    try:
        parsed = ast.literal_eval(value)
        if not isinstance(parsed, list) or not all(
            isinstance(x, int) for x in parsed
        ):
            raise ValueError("Input must be a list of Int, e.g., [5, 5, 1]")
        if len(parsed) != 3:
            raise ValueError(
                "PBC list must contain exactly 3 int, e.g., [5, 5, 1]"
            )
        return parsed
    except (ValueError, SyntaxError):
        raise argparse.ArgumentTypeError(f"Invalid int list: {value}")


def str2list(v: Optional[str]) -> Optional[list]:
    """Parse a string into a Python list via ast.literal_eval."""
    if v is None:
        return None
    try:
        return ast.literal_eval(v)
    except Exception:
        raise argparse.ArgumentTypeError(
            "k_path must be a valid Python list expression."
        )


if __name__ == "__main__":
    main()
